@@ -13,17 +13,24 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
+#include "config.h"
+
 #include <infinoted/infinoted-options.h>
+#include <infinoted/infinoted-util.h>
 #include <libinfinity/inf-i18n.h>
 
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+# include <libdaemon/dpid.h>
+#endif
+
 #include <glib.h>
 #include <stdlib.h>
 #include <string.h>
-
-#include "config.h"
+#include <errno.h>
 
 static const gchar INFINOTED_OPTIONS_GROUP[] = "infinoted";
 
@@ -31,9 +38,11 @@ static const gchar INFINOTED_OPTIONS_GROUP[] = "infinoted";
  * file. */
 
 /* We abuse the flags of a GOptionEntry to decide whether the option
- * can be set in the config file in addition to the command line */
-static const gint G_OPTION_FLAG_NO_CONFIG_FILE = 1 << 31;
+ * can be set in the config file in addition to the command line. This has to
+ * be a macro so that we can initialise the GOptionEntries[] with it. */
+#define G_OPTION_FLAG_NO_CONFIG_FILE (1 << 31)
 
+#if 0
 static gchar*
 infinoted_options_policy_to_string(InfXmppConnectionSecurityPolicy policy)
 {
@@ -51,6 +60,7 @@ infinoted_options_policy_to_string(InfXmppConnectionSecurityPolicy policy)
     return NULL;
   }
 }
+#endif
 
 static gboolean
 infinoted_options_policy_from_string(const gchar* string,
@@ -97,9 +107,9 @@ infinoted_options_port_to_integer(guint port)
 }
 
 static gboolean
-infinoted_options_autosave_interval_from_integer(gint value,
-                                                 guint* result,
-                                                 GError** error)
+infinoted_options_interval_from_integer(gint value,
+                                        guint* result,
+                                        GError** error)
 {
   if(value < 0)
   {
@@ -107,7 +117,8 @@ infinoted_options_autosave_interval_from_integer(gint value,
       error,
       infinoted_options_error_quark(),
       INFINOTED_OPTIONS_ERROR_INVALID_AUTOSAVE_INTERVAL,
-      _("Autosave interval must not be negative")
+      "%s",
+      _("Interval must not be negative")
     );
 
     return FALSE;
@@ -389,12 +400,21 @@ infinoted_options_validate(InfinotedOptions* options,
   InfXmppConnectionSecurityPolicy security_policy;
   security_policy = options->security_policy;
 
+  if(options->password != NULL &&
+     options->security_policy == INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED)
+  {
+    infinoted_util_log_warning(
+      _("Requiring password through unencrypted connection."));
+  }
+
+
   if(options->create_key == TRUE && options->create_certificate == FALSE)
   {
     g_set_error(
       error,
       infinoted_options_error_quark(),
       INFINOTED_OPTIONS_ERROR_INVALID_CREATE_OPTIONS,
+      "%s",
       _("Creating a new private key also requires creating a new certificate "
         "signed with it.")
     );
@@ -408,6 +428,7 @@ infinoted_options_validate(InfinotedOptions* options,
       error,
       infinoted_options_error_quark(),
       INFINOTED_OPTIONS_ERROR_EMPTY_KEY_FILE,
+      "%s",
       _("No private key file given. If you don't have a suitable key file, "
         "either create one using the --create-key command line argument, "
         "or disable TLS by setting the security policy to \"no-tls\".")
@@ -422,6 +443,7 @@ infinoted_options_validate(InfinotedOptions* options,
       error,
       infinoted_options_error_quark(),
       INFINOTED_OPTIONS_ERROR_EMPTY_CERTIFICATE_FILE,
+      "%s",
       _("No certificate file given. If you don't have a suitable certificate "
         "file, either create one using the --create-certificate command line "
         "agument, or disable TLS via by setting the security policy to "
@@ -430,6 +452,37 @@ infinoted_options_validate(InfinotedOptions* options,
 
     return FALSE;
   }
+  else if( (options->sync_directory != NULL && options->sync_interval == 0))
+  {
+    g_set_error(
+      error,
+      infinoted_options_error_quark(),
+      INFINOTED_OPTIONS_ERROR_INVALID_SYNC_COMBINATION,
+      "%s",
+      _("A synchronization directory is given, but synchronization interval "
+        "is not set. Please either set a nonzero synchronization interval "
+        "or unset the synchronization directory using the --sync-directory "
+        "and sync-interval command line or config file options.")
+    );
+
+    return FALSE;
+  }
+  else if(options->sync_directory == NULL && options->sync_interval != 0)
+  {
+    g_set_error(
+      error,
+      infinoted_options_error_quark(),
+      INFINOTED_OPTIONS_ERROR_INVALID_SYNC_COMBINATION,
+      "%s",
+      _("A synchronization interval is given, but the synchronization "
+        "directory is not set. Please either set a valid synchronization "
+        "directory, or set the synchronization interval to zero using "
+        "--sync-directory and sync-interval command line or config file "
+        "options.")
+    );
+
+    return FALSE;
+  }
 
   return TRUE;
 }
@@ -446,54 +499,103 @@ infinoted_options_load(InfinotedOptions* options,
   gchar* security_policy;
   gint port_number;
   gboolean display_version;
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+  gboolean kill_daemon;
+#endif
   gint autosave_interval;
+  gint sync_interval;
+  guint i;
 
   gboolean result;
 
   GOptionContext *context;
   gchar* desc;
 
-  const GOptionEntry entries[] = {
+  GOptionEntry entries[] = {
     { "key-file", 'k', 0,
-      G_OPTION_ARG_FILENAME, &options->key_file,
+      G_OPTION_ARG_FILENAME, NULL,
       N_("The server's private key"), N_("KEY-FILE") },
     { "certificate-file", 'c', 0,
-      G_OPTION_ARG_FILENAME, &options->certificate_file,
+      G_OPTION_ARG_FILENAME, NULL,
       N_("The server's certificate"), N_("CERTIFICATE-FILE") },
     { "certificate-chain", 0, 0,
-      G_OPTION_ARG_FILENAME, &options->certificate_chain_file,
+      G_OPTION_ARG_FILENAME, NULL,
       N_("The certificates chain down to the root certificate"), NULL },
     { "create-key", 0, G_OPTION_FLAG_NO_CONFIG_FILE,
-      G_OPTION_ARG_NONE, &options->create_key,
+      G_OPTION_ARG_NONE, NULL,
       N_("Creates a new random private key"), NULL },
     { "create-certificate", 0, G_OPTION_FLAG_NO_CONFIG_FILE,
-      G_OPTION_ARG_NONE, &options->create_certificate,
+      G_OPTION_ARG_NONE, NULL,
       N_("Creates a new self-signed certificate using the given key"), NULL },
     { "port-number", 'p', 0,
-      G_OPTION_ARG_INT, &port_number,
+      G_OPTION_ARG_INT, NULL,
       N_("The port number to listen on"), N_("PORT") },
     { "security-policy", 0, 0,
-      G_OPTION_ARG_STRING, &security_policy,
+      G_OPTION_ARG_STRING, NULL,
       N_("How to decide whether to use TLS"), "no-tls|allow-tls|require-tls" },
     { "root-directory", 'r', 0,
-      G_OPTION_ARG_FILENAME, &options->root_directory,
+      G_OPTION_ARG_FILENAME, NULL,
       N_("The directory to store documents into"), N_("DIRECTORY") },
     { "autosave-interval", 0, 0,
-      G_OPTION_ARG_INT, &autosave_interval,
+      G_OPTION_ARG_INT, NULL,
       N_("Interval within which to save documents, in seconds, or 0 to "
          "disable autosave"), N_("INTERVAL") },
+    { "password", 'P', 0,
+      G_OPTION_ARG_STRING, NULL,
+      N_("Require given password on connections"), N_("PASSWORD") },
+    { "sync-directory", 0, 0,
+      G_OPTION_ARG_FILENAME, NULL,
+      N_("A directory into which to periodically store a copy of the "
+         "document tree"), N_("DIRECTORY") },
+    { "sync-interval", 0, 0,
+      G_OPTION_ARG_INT, NULL,
+      N_("Interval within which to store documents to the specified "
+         "sync-directory, or 0 to disable directory synchronization"),
+         N_("INTERVAL") },
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+    { "daemonize", 'd', 0,
+      G_OPTION_ARG_NONE, NULL,
+      N_("Daemonize the server"), NULL },
+    { "kill-daemon", 'D', 0,
+      G_OPTION_ARG_NONE, NULL,
+      N_("Kill a running daemon"), NULL },
+#endif
     { "version", 'v', G_OPTION_FLAG_NO_CONFIG_FILE,
-      G_OPTION_ARG_NONE, &display_version,
+      G_OPTION_ARG_NONE, NULL,
       N_("Display version information and exit"), NULL },
     { NULL, 0, 0, G_OPTION_ARG_NONE,
       NULL,
       NULL, 0 }
   };
 
+  /* C90 does not allow non-compile-time-constant initializers for structs */
+  i = 0;
+  entries[i++].arg_data = &options->key_file;
+  entries[i++].arg_data = &options->certificate_file;
+  entries[i++].arg_data = &options->certificate_chain_file;
+  entries[i++].arg_data = &options->create_key;
+  entries[i++].arg_data = &options->create_certificate;
+  entries[i++].arg_data = &port_number;
+  entries[i++].arg_data = &security_policy;
+  entries[i++].arg_data = &options->root_directory;
+  entries[i++].arg_data = &autosave_interval;
+  entries[i++].arg_data = &options->password;
+  entries[i++].arg_data = &options->sync_directory;
+  entries[i++].arg_data = &sync_interval;
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+  entries[i++].arg_data = &options->daemonize;
+  entries[i++].arg_data = &kill_daemon;
+#endif
+  entries[i++].arg_data = &display_version;
+
   display_version = FALSE;
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+  kill_daemon = FALSE;
+#endif
   security_policy = NULL;
   port_number = infinoted_options_port_to_integer(options->port);
   autosave_interval = options->autosave_interval;
+  sync_interval = options->sync_interval;
 
   if(config_files)
   {
@@ -508,26 +610,52 @@ infinoted_options_load(InfinotedOptions* options,
     }
   }
 
-  desc = g_strdup_printf("- %s", _("infinote dedicated server"));
-  context = g_option_context_new(desc);
-  g_free(desc);
-  g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
-
-  if(g_option_context_parse(context, argc, argv, error) == FALSE)
+  if(argc != NULL && argv != NULL)
   {
+    desc = g_strdup_printf("- %s", _("infinote dedicated server"));
+    context = g_option_context_new(desc);
+    g_free(desc);
+    g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
+
+    if(g_option_context_parse(context, argc, argv, error) == FALSE)
+    {
+      g_option_context_free(context);
+      g_free(security_policy);
+      return FALSE;
+    }
+
+    if(display_version)
+    {
+      printf("infinoted %s\n", PACKAGE_VERSION);
+      exit(0);
+    }
+
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+    if(kill_daemon)
+    {
+      g_free(security_policy);
+
+      if(!infinoted_util_set_daemon_pid_file_proc(error))
+        return FALSE;
+
+#ifdef DAEMON_PID_FILE_KILL_WAIT_AVAILABLE
+      if(daemon_pid_file_kill_wait(SIGTERM, 5) != 0)
+#else
+      if(daemon_pid_file_kill(SIGTERM) != 0)
+#endif
+      {
+        infinoted_util_set_errno_error(error, errno,
+          _("Could not kill daemon"));
+        return FALSE;
+      }
+
+      exit(0);
+    }
+#endif
+
     g_option_context_free(context);
-    g_free(security_policy);
-    return FALSE;
   }
 
-  if(display_version)
-  {
-    printf("infinoted %s\n", PACKAGE_VERSION);
-    exit(0);
-  }
-
-  g_option_context_free(context);
-
   if(security_policy != NULL)
   {
     result = infinoted_options_policy_from_string(
@@ -540,6 +668,8 @@ infinoted_options_load(InfinotedOptions* options,
     if(!result) return FALSE;
   }
 
+  /* TODO: Do we leak security_policy at this point? */
+
   result = infinoted_options_port_from_integer(
     port_number,
     &options->port,
@@ -547,13 +677,33 @@ infinoted_options_load(InfinotedOptions* options,
   );
   if(!result) return FALSE;
 
-  result = infinoted_options_autosave_interval_from_integer(
+  result = infinoted_options_interval_from_integer(
     autosave_interval,
     &options->autosave_interval,
     error
   );
   if(!result) return FALSE;
 
+  result = infinoted_options_interval_from_integer(
+    sync_interval,
+    &options->sync_interval,
+    error
+  );
+  if(!result) return FALSE;
+
+  if(options->password != NULL && strcmp(options->password, "") == 0)
+  {
+    g_free(options->password);
+    options->password = NULL;
+  }
+
+  if(options->sync_directory != NULL &&
+     strcmp(options->sync_directory, "") == 0)
+  {
+    g_free(options->sync_directory);
+    options->sync_directory = NULL;
+  }
+
   return infinoted_options_validate(options, error);
 }
 
@@ -597,6 +747,13 @@ infinoted_options_new(const gchar* const* config_files,
   options->root_directory =
     g_build_filename(g_get_home_dir(), ".infinote", NULL);
   options->autosave_interval = 0;
+  options->password = NULL;
+  options->sync_directory = NULL;
+  options->sync_interval = 0;
+
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+  options->daemonize = FALSE;
+#endif
 
   if(!infinoted_options_load(options, config_files, argc, argv, error))
   {
@@ -620,6 +777,8 @@ infinoted_options_free(InfinotedOptions* options)
   g_free(options->certificate_file);
   g_free(options->certificate_chain_file);
   g_free(options->root_directory);
+  g_free(options->password);
+  g_free(options->sync_directory);
   g_slice_free(InfinotedOptions, options);
 }
 
