@@ -13,13 +13,15 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #ifndef __INFINOTED_SIGNAL_H__
 #define __INFINOTED_SIGNAL_H__
 
 #include <infinoted/infinoted-run.h>
+#include <libinfinity/inf-config.h>
 
 #include <glib.h>
 
@@ -32,8 +34,15 @@ typedef void(*InfinotedSignalFunc)(int);
 
 typedef struct _InfinotedSignal InfinotedSignal;
 struct _InfinotedSignal {
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+  InfinotedRun* run;
+  int signal_fd;
+#else
   InfinotedSignalFunc previous_sigint_handler;
   InfinotedSignalFunc previous_sigterm_handler;
+  InfinotedSignalFunc previous_sigquit_handler;
+  InfinotedSignalFunc previous_sighup_handler;
+#endif
 };
 
 InfinotedSignal*
