@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfgtk/inf-gtk-browser-store.h>
@@ -21,6 +22,7 @@
 #include <libinfinity/client/infc-browser.h>
 #include <libinfinity/inf-marshal.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <gtk/gtktreemodel.h>
 
@@ -41,12 +43,14 @@ struct _InfGtkBrowserStoreItem {
   InfDiscovery* discovery;
   InfDiscoveryInfo* info;
 
- /* This is the same as infc_browser_get_connection, but we need an extra
-  * reference on this because we connect to its "error" signal and need to
-  * disconnect when the browser already released its reference: */
-  InfXmlConnection* connection;
   InfcBrowser* browser;
 
+  /* Browser node which is currently to be removed. This is required since
+   * when node-removed is emitted in InfcBrowser the node is still present,
+   * but we need to make sure that the GtkTreeModel functions do as if it
+   * wasn't present anymore. */
+  gpointer missing;
+
   /* Running requests */
   GSList* requests;
   /* Saved node errors (during exploration/subscription) */
@@ -105,8 +109,8 @@ inf_gtk_browser_store_find_item_by_connection(InfGtkBrowserStore* store,
   priv = INF_GTK_BROWSER_STORE_PRIVATE(store);
 
   for(item = priv->first_item; item != NULL; item = item->next)
-    if(item->connection != NULL)
-      if(item->connection == connection)
+    if(item->browser != NULL)
+      if(infc_browser_get_connection(item->browser) == connection)
         return item;
 
   return NULL;
@@ -151,14 +155,14 @@ inf_gtk_browser_store_find_item_by_discovery_info(InfGtkBrowserStore* store,
  */
 
 static void
-inf_gtk_browser_store_connection_notify_status_cb(GObject* object,
-                                                  GParamSpec* pspec,
-                                                  gpointer user_data);
+inf_gtk_browser_store_browser_notify_status_cb(GObject* object,
+                                               GParamSpec* pspec,
+                                               gpointer user_data);
 
 static void
-inf_gtk_browser_store_connection_error_cb(InfXmlConnection* connection,
-                                          const GError* error,
-                                          gpointer user_data);
+inf_gtk_browser_store_browser_error_cb(InfcBrowser* browser,
+                                       const GError* error,
+                                       gpointer user_data);
 
 static void
 inf_gtk_browser_store_node_added_cb(InfcBrowser* browser,
@@ -212,7 +216,7 @@ inf_gtk_browser_store_item_request_remove(InfGtkBrowserStoreItem* item,
     item
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(request),
     G_CALLBACK(inf_gtk_browser_store_request_failed_cb),
     item
@@ -381,9 +385,11 @@ inf_gtk_browser_store_add_item(InfGtkBrowserStore* store,
   item->name = name;
   item->discovery = discovery;
   item->info = info;
-  item->status = INF_GTK_BROWSER_MODEL_DISCOVERED;
+  item->status = INF_GTK_BROWSER_MODEL_DISCONNECTED;
+  if(discovery != NULL && info != NULL)
+    item->status = INF_GTK_BROWSER_MODEL_DISCOVERED;
   item->browser = NULL;
-  item->connection = NULL;
+  item->missing = NULL;
   item->node_errors = g_hash_table_new_full(
     NULL,
     NULL,
@@ -491,7 +497,7 @@ inf_gtk_browser_store_remove_item(InfGtkBrowserStore* store,
 
   gtk_tree_model_row_deleted(GTK_TREE_MODEL(store), path);
   gtk_tree_path_free(path);
-  
+
   if(item->error != NULL)
     g_error_free(item->error);
 
@@ -528,16 +534,31 @@ inf_gtk_browser_store_undiscovered_cb(InfDiscovery* discovery,
 
   store = INF_GTK_BROWSER_STORE(user_data);
   item = inf_gtk_browser_store_find_item_by_discovery_info(store, info);
-  g_assert(item != NULL);
 
-  /* TODO: Keep if connection exists, just reset discovery and info */
-  inf_gtk_browser_store_remove_item(store, item);
+  /* If, after a discovery resove we see that there is already an item for
+   * that connection, then we remove the discovery item, and use the other
+   * item instead. In that case, item can be NULL here. */
+  if(item != NULL)
+  {
+    if(item->browser == NULL)
+    {
+      inf_gtk_browser_store_remove_item(store, item);
+    }
+    else
+    {
+      g_assert(item->status != INF_GTK_BROWSER_MODEL_DISCOVERED &&
+               item->status != INF_GTK_BROWSER_MODEL_RESOLVING);
+
+      item->discovery = NULL;
+      item->info = NULL;
+    }
+  }
 }
 
 static void
-inf_gtk_browser_store_connection_error_cb(InfXmlConnection* connection,
-                                          const GError* error,
-                                          gpointer user_data)
+inf_gtk_browser_store_browser_error_cb(InfcBrowser* browser,
+                                       const GError* error,
+                                       gpointer user_data)
 {
   InfGtkBrowserStore* store;
   InfGtkBrowserStoreItem* item;
@@ -547,7 +568,7 @@ inf_gtk_browser_store_connection_error_cb(InfXmlConnection* connection,
 
   store = INF_GTK_BROWSER_STORE(user_data);
   priv = INF_GTK_BROWSER_STORE_PRIVATE(store);
-  item = inf_gtk_browser_store_find_item_by_connection(store, connection);
+  item = inf_gtk_browser_store_find_item_by_browser(store, browser);
   g_assert(item != NULL);
 
   /* Overwrite previous error */
@@ -569,25 +590,23 @@ inf_gtk_browser_store_connection_error_cb(InfXmlConnection* connection,
 }
 
 static void
-inf_gtk_browser_store_connection_notify_status_cb(GObject* object,
-                                                  GParamSpec* pspec,
-                                                  gpointer user_data)
+inf_gtk_browser_store_browser_notify_status_cb(GObject* object,
+                                               GParamSpec* pspec,
+                                               gpointer user_data)
 {
   InfGtkBrowserStore* store;
   InfGtkBrowserStorePrivate* priv;
-  InfXmlConnection* connection;
+  InfcBrowser* browser;
   InfGtkBrowserStoreItem* item;
-  InfXmlConnectionStatus status;
   GtkTreeIter iter;
   GtkTreePath* path;
 
   store = INF_GTK_BROWSER_STORE(user_data);
   priv = INF_GTK_BROWSER_STORE_PRIVATE(user_data);  
-  connection = INF_XML_CONNECTION(object);
-  item = inf_gtk_browser_store_find_item_by_connection(store, connection);
+  browser = INFC_BROWSER(object);
+  item = inf_gtk_browser_store_find_item_by_browser(store, browser);
 
   g_assert(item != NULL);
-  g_assert(item->status != INF_GTK_BROWSER_MODEL_ERROR);
 
   iter.stamp = priv->stamp;
   iter.user_data = item;
@@ -596,20 +615,10 @@ inf_gtk_browser_store_connection_notify_status_cb(GObject* object,
 
   path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
 
-  g_object_get(G_OBJECT(connection), "status", &status, NULL);
-
-  switch(status)
+  switch(infc_browser_get_status(browser))
   {
-  case INF_XML_CONNECTION_OPENING:
-    item->status = INF_GTK_BROWSER_MODEL_CONNECTING;
-    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
-    break;
-  case INF_XML_CONNECTION_OPEN:
-    item->status = INF_GTK_BROWSER_MODEL_CONNECTED;
-    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
-    break;
-  case INF_XML_CONNECTION_CLOSING:
-  case INF_XML_CONNECTION_CLOSED:
+  case INFC_BROWSER_DISCONNECTED:
+    /* TODO: Do we want to go to disconnected state when error is not set? */
     item->status = INF_GTK_BROWSER_MODEL_ERROR;
 
     /* Set a "Disconnected" error if there is not already one set by
@@ -617,23 +626,26 @@ inf_gtk_browser_store_connection_notify_status_cb(GObject* object,
      * meaningful error message. */
     if(item->error == NULL)
     {
-      g_set_error(
-        &item->error,
+      item->error = g_error_new_literal(
         g_quark_from_static_string("INF_GTK_BROWSER_STORE_ERROR"),
         0,
         _("Disconnected")
       );
     }
 
-    /* set_browser() will do this anyway, in the default handler: */
-    /*gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);*/
-
-    /* Reset browser, we do not need it anymore since its connection is
-     * gone. */
-    /* TODO: Keep browser and still allow browsing in explored folders,
-     * but reset connection. */
-    /* Status cannot change to invalid because we just set error */
-    inf_gtk_browser_store_item_set_browser(store, item, path, NULL);
+    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
+    break;
+  case INFC_BROWSER_CONNECTING:
+    if(item->error != NULL) g_error_free(item->error);
+    item->error = NULL;
+    item->status = INF_GTK_BROWSER_MODEL_CONNECTING;
+    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
+    break;
+  case INFC_BROWSER_CONNECTED:
+    if(item->error != NULL) g_error_free(item->error);
+    item->error = NULL;
+    item->status = INF_GTK_BROWSER_MODEL_CONNECTED;
+    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
     break;
   default:
     g_assert_not_reached();
@@ -718,12 +730,19 @@ inf_gtk_browser_store_node_removed_cb(InfcBrowser* browser,
   priv = INF_GTK_BROWSER_STORE_PRIVATE(store);
   item = inf_gtk_browser_store_find_item_by_browser(store, browser);
 
+  g_assert(item->missing == NULL);
+
   tree_iter.stamp = priv->stamp;
   tree_iter.user_data = item;
   tree_iter.user_data2 = GUINT_TO_POINTER(iter->node_id);
   tree_iter.user_data3 = iter->node;
 
   path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &tree_iter);
+
+  /* This is a small hack to have the item removed from the tree
+   * model before it is removed from the InfcBrowser. */
+  item->missing = iter->node;
+
   gtk_tree_model_row_deleted(GTK_TREE_MODEL(store), path);
 
   /* TODO: Remove requests and node errors from nodes below the removed one */
@@ -759,6 +778,7 @@ inf_gtk_browser_store_node_removed_cb(InfcBrowser* browser,
   }
 
   gtk_tree_path_free(path);
+  item->missing = NULL;
 }
 
 static void
@@ -786,9 +806,18 @@ inf_gtk_browser_store_begin_subscribe_cb(InfcBrowser* browser,
   InfGtkBrowserStoreItem* item;
 
   store = INF_GTK_BROWSER_STORE(user_data);
-  item = inf_gtk_browser_store_find_item_by_browser(store, browser);
 
-  inf_gtk_browser_store_item_request_add(store, item, INFC_REQUEST(request));
+  /* Ignore chat subscriptions */
+  if(iter != NULL)
+  {
+    item = inf_gtk_browser_store_find_item_by_browser(store, browser);
+
+    inf_gtk_browser_store_item_request_add(
+      store,
+      item,
+      INFC_REQUEST(request)
+    );
+  }
 }
 
 static void
@@ -818,7 +847,7 @@ inf_gtk_browser_store_resolv_complete_func(InfDiscoveryInfo* info,
   priv = INF_GTK_BROWSER_STORE_PRIVATE(store);
   new_item = inf_gtk_browser_store_find_item_by_discovery_info(store, info);
   old_item = inf_gtk_browser_store_find_item_by_connection(store, connection);
-  
+
   g_assert(new_item != NULL);
   g_assert(new_item->status == INF_GTK_BROWSER_MODEL_RESOLVING);
 
@@ -903,7 +932,8 @@ inf_gtk_browser_store_resolv_complete_func(InfDiscoveryInfo* info,
       gtk_tree_path_free(path);
 
       /* TODO: Perhaps we should emit a signal so that the view can
-       * highlight and scroll to the existing item. */
+       * highlight and scroll to the existing item. And also so that it can
+       * over initial root exploration. */
 
       g_free(order);
     }
@@ -997,13 +1027,13 @@ inf_gtk_browser_store_dispose(GObject* object)
 
   for(item = priv->discoveries; item != NULL; item = g_slist_next(item))
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->data),
       G_CALLBACK(inf_gtk_browser_store_discovered_cb),
       store
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->data),
       G_CALLBACK(inf_gtk_browser_store_undiscovered_cb),
       store
@@ -1052,7 +1082,7 @@ inf_gtk_browser_store_set_property(GObject* object,
     g_assert(priv->communication_manager == NULL); /* construct only */
     priv->communication_manager =
       INF_COMMUNICATION_MANAGER(g_value_dup_object(value));
-  
+
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -1116,6 +1146,8 @@ inf_gtk_browser_store_tree_model_get_column_type(GtkTreeModel* model,
     return INFC_TYPE_BROWSER;
   case INF_GTK_BROWSER_MODEL_COL_STATUS:
     return INF_GTK_TYPE_BROWSER_MODEL_STATUS;
+  case INF_GTK_BROWSER_MODEL_COL_NAME:
+    return G_TYPE_STRING;
   case INF_GTK_BROWSER_MODEL_COL_ERROR:
     return G_TYPE_POINTER;
   case INF_GTK_BROWSER_MODEL_COL_NODE:
@@ -1147,7 +1179,7 @@ inf_gtk_browser_store_tree_model_get_iter(GtkTreeModel* model,
 
   i = 0;
   for(item = priv->first_item; item != NULL && i < n; item = item->next)
-    ++ i;
+    ++i;
 
   if(item == NULL) return FALSE;
 
@@ -1163,7 +1195,7 @@ inf_gtk_browser_store_tree_model_get_iter(GtkTreeModel* model,
 
   if(item->browser == NULL) return FALSE;
   infc_browser_iter_get_root(item->browser, &browser_iter);
-  
+
   for(n = 1; n < (guint)gtk_tree_path_get_depth(path); ++ n)
   {
     if(infc_browser_iter_get_explored(item->browser, &browser_iter) == FALSE)
@@ -1172,10 +1204,18 @@ inf_gtk_browser_store_tree_model_get_iter(GtkTreeModel* model,
     if(infc_browser_iter_get_child(item->browser, &browser_iter) == FALSE)
       return FALSE;
 
+    /* skip missing */
+    if(browser_iter.node == item->missing)
+      ++indices[n];
+
     for(i = 0; i < (guint)indices[n]; ++ i)
     {
       if(infc_browser_iter_get_next(item->browser, &browser_iter) == FALSE)
         return FALSE;
+
+      /* skip missing */
+      if(browser_iter.node == item->missing)
+        ++indices[n];
     }
   }
 
@@ -1199,7 +1239,7 @@ inf_gtk_browser_store_tree_model_get_path_impl(InfGtkBrowserStore* store,
   InfGtkBrowserStoreItem* cur;
   gboolean result;
   guint n;
-  
+
   cur_iter = *iter;
   if(infc_browser_iter_get_parent(item->browser, &cur_iter) == FALSE)
   {
@@ -1208,7 +1248,7 @@ inf_gtk_browser_store_tree_model_get_path_impl(InfGtkBrowserStore* store,
     /* We are on top level, but still need to find the item index */
     n = 0;
     for(cur = priv->first_item; cur != item; cur = cur->next)
-      ++ n;
+      ++n;
 
     gtk_tree_path_append_index(path, n);
   }
@@ -1224,12 +1264,27 @@ inf_gtk_browser_store_tree_model_get_path_impl(InfGtkBrowserStore* store,
     result = infc_browser_iter_get_child(item->browser, &cur_iter);
     g_assert(result == TRUE);
 
+    /* skip missing */
+    if(cur_iter.node == item->missing)
+    {
+      result = infc_browser_iter_get_next(item->browser, &cur_iter);
+      g_assert(result == TRUE);
+    }
+
     n = 0;
     while(cur_iter.node_id != iter->node_id)
     {
       result = infc_browser_iter_get_next(item->browser, &cur_iter);
       g_assert(result == TRUE);
-      ++ n;
+
+      /* skip missing */
+      if(cur_iter.node == item->missing)
+      {
+        result = infc_browser_iter_get_next(item->browser, &cur_iter);
+        g_assert(result == TRUE);
+      }
+
+      ++n;
     }
 
     gtk_tree_path_append_index(path, n);
@@ -1250,7 +1305,7 @@ inf_gtk_browser_store_tree_model_get_path(GtkTreeModel* model,
   priv = INF_GTK_BROWSER_STORE_PRIVATE(model);
   g_assert(iter->stamp == priv->stamp);
   g_assert(iter->user_data != NULL);
-  
+
   item = (InfGtkBrowserStoreItem*)iter->user_data;
 
   path = gtk_tree_path_new();
@@ -1273,7 +1328,7 @@ inf_gtk_browser_store_tree_model_get_path(GtkTreeModel* model,
     /* toplevel */
     n = 0;
     for(cur = priv->first_item; cur != item; cur = cur->next)
-      ++ n;
+      ++n;
 
     gtk_tree_path_append_index(path, n);
   }
@@ -1298,7 +1353,8 @@ inf_gtk_browser_store_tree_model_get_value(GtkTreeModel* model,
   item = (InfGtkBrowserStoreItem*)iter->user_data;
   browser_iter.node_id = GPOINTER_TO_UINT(iter->user_data2);
   browser_iter.node = iter->user_data3;
-  
+  g_assert(item->missing == NULL || browser_iter.node != item->missing);
+
   switch(column)
   {
   case INF_GTK_BROWSER_MODEL_COL_DISCOVERY_INFO:
@@ -1411,9 +1467,16 @@ inf_gtk_browser_store_tree_model_iter_next(GtkTreeModel* model,
   }
   else
   {
+    g_assert(browser_iter.node != item->missing);
+
     if(infc_browser_iter_get_next(item->browser, &browser_iter) == FALSE)
       return FALSE;
 
+    /* skip missing */
+    if(browser_iter.node == item->missing)
+      if(infc_browser_iter_get_next(item->browser, &browser_iter) == FALSE)
+        return FALSE;
+
     iter->user_data2 = GUINT_TO_POINTER(browser_iter.node_id);
     iter->user_data3 = browser_iter.node;
     return TRUE;
@@ -1456,6 +1519,8 @@ inf_gtk_browser_store_tree_model_iter_children(GtkTreeModel* model,
     else
       browser_iter.node = parent->user_data3;
 
+    g_assert(item->missing == NULL || browser_iter.node != item->missing);
+
     if(!infc_browser_iter_is_subdirectory(item->browser, &browser_iter))
       return FALSE;
 
@@ -1465,6 +1530,10 @@ inf_gtk_browser_store_tree_model_iter_children(GtkTreeModel* model,
     if(!infc_browser_iter_get_child(item->browser, &browser_iter))
       return FALSE;
 
+    if(browser_iter.node == item->missing)
+      if(!infc_browser_iter_get_next(item->browser, &browser_iter))
+        return FALSE;
+
     iter->stamp = priv->stamp;
     iter->user_data = item;
     iter->user_data2 = GUINT_TO_POINTER(browser_iter.node_id);
@@ -1489,6 +1558,7 @@ inf_gtk_browser_store_tree_model_iter_has_child(GtkTreeModel* model,
 
   browser_iter.node_id = GPOINTER_TO_UINT(iter->user_data2);
   browser_iter.node = iter->user_data3;
+  g_assert(item->missing == NULL || browser_iter.node != item->missing);
 
   if(browser_iter.node == NULL)
     infc_browser_iter_get_root(item->browser, &browser_iter);
@@ -1499,7 +1569,14 @@ inf_gtk_browser_store_tree_model_iter_has_child(GtkTreeModel* model,
   if(infc_browser_iter_get_explored(item->browser, &browser_iter) == FALSE)
     return FALSE;
 
-  return infc_browser_iter_get_child(item->browser, &browser_iter);
+  if(!infc_browser_iter_get_child(item->browser, &browser_iter))
+    return FALSE;
+
+  if(browser_iter.node == item->missing)
+    if(!infc_browser_iter_get_next(item->browser, &browser_iter))
+      return FALSE;
+
+  return TRUE;
 }
 
 static gint
@@ -1520,7 +1597,7 @@ inf_gtk_browser_store_tree_model_iter_n_children(GtkTreeModel* model,
   {
     n = 0;
     for(cur = priv->first_item; cur != NULL; cur = cur->next)
-      ++ n;
+      ++n;
 
     return n;
   }
@@ -1529,6 +1606,7 @@ inf_gtk_browser_store_tree_model_iter_n_children(GtkTreeModel* model,
     item = (InfGtkBrowserStoreItem*)iter->user_data;
     browser_iter.node_id = GPOINTER_TO_UINT(iter->user_data2);
     browser_iter.node = iter->user_data3;
+    g_assert(item->missing == NULL || browser_iter.node != item->missing);
 
     if(browser_iter.node == NULL)
       infc_browser_iter_get_root(item->browser, &browser_iter);
@@ -1541,7 +1619,8 @@ inf_gtk_browser_store_tree_model_iter_n_children(GtkTreeModel* model,
         result == TRUE;
         result = infc_browser_iter_get_next(item->browser, &browser_iter))
     {
-      ++ n;
+      if(browser_iter.node != item->missing)
+        ++n;
     }
 
     return n;
@@ -1590,17 +1669,25 @@ inf_gtk_browser_store_tree_model_iter_nth_child(GtkTreeModel* model,
       infc_browser_iter_get_root(item->browser, &browser_iter);
     else
       browser_iter.node = parent->user_data3;
+    g_assert(item->missing == NULL || browser_iter.node != item->missing);
 
     if(infc_browser_iter_get_explored(item->browser, &browser_iter) == FALSE)
       return FALSE;
-                
+
     if(infc_browser_iter_get_child(item->browser, &browser_iter) == FALSE)
       return FALSE;
 
+    /* skip missing */
+    if(browser_iter.node == item->missing)
+      ++n;
+
     for(i = 0; i < (guint)n; ++ i)
     {
       if(infc_browser_iter_get_next(item->browser, &browser_iter) == FALSE)
         return FALSE;
+
+      if(browser_iter.node == item->missing)
+        ++n;
     }
 
     iter->stamp = priv->stamp;
@@ -1631,8 +1718,11 @@ inf_gtk_browser_store_tree_model_iter_parent(GtkTreeModel* model,
   if(browser_iter.node == NULL)
     return FALSE;
 
+  g_assert(browser_iter.node != item->missing);
+
   result = infc_browser_iter_get_parent(item->browser, &browser_iter);
   g_assert(result == TRUE);
+  g_assert(browser_iter.node != item->missing);
 
   iter->stamp = priv->stamp;
   iter->user_data = item;
@@ -1658,7 +1748,6 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
 {
   InfGtkBrowserStorePrivate* priv;
   InfGtkBrowserStoreItem* item;
-  InfXmlConnectionStatus status;
 
   InfcBrowserIter iter;
   guint n;
@@ -1696,23 +1785,6 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
       gtk_tree_path_up(path);
     }
 
-    if(item->connection != NULL)
-    {
-      g_signal_handlers_disconnect_by_func(
-        G_OBJECT(item->connection),
-        G_CALLBACK(inf_gtk_browser_store_connection_error_cb),
-        model
-      );
-
-      g_signal_handlers_disconnect_by_func(
-        G_OBJECT(item->connection),
-        G_CALLBACK(inf_gtk_browser_store_connection_notify_status_cb),
-        model
-      );
-
-      g_object_unref(G_OBJECT(item->connection));
-    }
-
     while(item->requests != NULL)
     {
       inf_gtk_browser_store_item_request_remove(
@@ -1723,25 +1795,37 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
 
     g_hash_table_remove_all(item->node_errors);
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
+      G_OBJECT(item->browser),
+      G_CALLBACK(inf_gtk_browser_store_browser_error_cb),
+      model
+    );
+
+    inf_signal_handlers_disconnect_by_func(
+      G_OBJECT(item->browser),
+      G_CALLBACK(inf_gtk_browser_store_browser_notify_status_cb),
+      model
+    );
+
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->browser),
       G_CALLBACK(inf_gtk_browser_store_node_added_cb),
       model
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->browser),
       G_CALLBACK(inf_gtk_browser_store_node_removed_cb),
       model
     );
-    
-    g_signal_handlers_disconnect_by_func(
+
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->browser),
       G_CALLBACK(inf_gtk_browser_store_begin_explore_cb),
       model
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(item->browser),
       G_CALLBACK(inf_gtk_browser_store_begin_subscribe_cb),
       model
@@ -1761,35 +1845,26 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
     );
   }
 
-  /* Set up new browser and connection */
+  /* Set up new browser */
   item->browser = browser;
-  if(browser != NULL)
-    item->connection = infc_browser_get_connection(browser);
-  else
-    item->connection = NULL;
 
   if(browser != NULL)
   {
-    if(item->connection != NULL)
-    {
-      g_object_ref(G_OBJECT(item->connection));
+    g_object_ref(browser);
 
-      g_signal_connect(
-        G_OBJECT(item->connection),
-        "error",
-        G_CALLBACK(inf_gtk_browser_store_connection_error_cb),
-        model
-      );
+    g_signal_connect(
+      G_OBJECT(item->browser),
+      "error",
+      G_CALLBACK(inf_gtk_browser_store_browser_error_cb),
+      model
+    );
       
-      g_signal_connect(
-        G_OBJECT(item->connection),
-        "notify::status",
-        G_CALLBACK(inf_gtk_browser_store_connection_notify_status_cb),
-        model
-      );
-    }
-
-    g_object_ref(G_OBJECT(browser));
+    g_signal_connect(
+      G_OBJECT(item->browser),
+      "notify::status",
+      G_CALLBACK(inf_gtk_browser_store_browser_notify_status_cb),
+      model
+    );
 
     g_signal_connect_after(
       G_OBJECT(item->browser),
@@ -1827,25 +1902,25 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
   if(item->browser == NULL && item->info == NULL && item->error == NULL)
   {
     item->status = INF_GTK_BROWSER_MODEL_INVALID;
+    /* TODO: What's up with this? Can this happen? Can we assert() here? */
   }
   else if(item->status != INF_GTK_BROWSER_MODEL_ERROR)
   {
-    /* Set item status according to connection status if there is no error
-     * set. */
-    if(item->connection != NULL)
+    /* Set item status according to browser status if there is no
+     * error set. */
+    if(item->browser != NULL)
     {
-      g_object_get(G_OBJECT(item->connection), "status", &status, NULL);
-      switch(status)
+      switch(infc_browser_get_status(item->browser))
       {
-      case INF_XML_CONNECTION_OPENING:
+      case INFC_BROWSER_DISCONNECTED:
+        item->status = INF_GTK_BROWSER_MODEL_DISCONNECTED;
+        break;
+      case INFC_BROWSER_CONNECTING:
         item->status = INF_GTK_BROWSER_MODEL_CONNECTING;
         break;
-      case INF_XML_CONNECTION_OPEN:
+      case INFC_BROWSER_CONNECTED:
         item->status = INF_GTK_BROWSER_MODEL_CONNECTED;
         break;
-      case INF_XML_CONNECTION_CLOSING:
-      case INF_XML_CONNECTION_CLOSED:
-        /* Browser drops connection if in one of those states */
       default:
         g_assert_not_reached();
         break;
@@ -1853,7 +1928,7 @@ inf_gtk_browser_store_browser_model_set_browser(InfGtkBrowserModel* model,
     }
     else
     {
-      /* No connection available. Discovery needs to be set now, otherwise
+      /* No browser available. Discovery needs to be set now, otherwise
        * we would have set the status to invalid above. */
       g_assert(item->info != NULL);
       item->status = INF_GTK_BROWSER_MODEL_DISCOVERED;
@@ -1892,7 +1967,7 @@ inf_gtk_browser_store_browser_model_resolve(InfGtkBrowserModel* model,
   g_assert(
     item->status == INF_GTK_BROWSER_MODEL_DISCOVERED ||
     item->status == INF_GTK_BROWSER_MODEL_ERROR
-  );    
+  );
 
   if(item->status == INF_GTK_BROWSER_MODEL_ERROR)
   {
@@ -1941,7 +2016,7 @@ inf_gtk_browser_store_browser_iter_to_tree_iter(InfGtkBrowserModel* model,
     INF_GTK_BROWSER_STORE(model),
     browser
   );
-  if(item == NULL) return FALSE;
+  if(item == NULL || item->missing == browser_iter->node) return FALSE;
 
   tree_iter->stamp = priv->stamp;
   tree_iter->user_data = item;
@@ -2191,9 +2266,6 @@ inf_gtk_browser_store_add_discovery(InfGtkBrowserStore* store,
  * browse the explored parts of the directory of the remote site. If @name
  * is %NULL, then the #InfXmlConnection:remote-id of the connection will be
  * used.
- *
- * @connection must be in %INF_XML_CONNECTION_OPEN or
- * %INF_XML_CONNECTION_OPENING status.
  **/
 void
 inf_gtk_browser_store_add_connection(InfGtkBrowserStore* store,
@@ -2209,8 +2281,6 @@ inf_gtk_browser_store_add_connection(InfGtkBrowserStore* store,
   g_return_if_fail(INF_IS_XML_CONNECTION(connection));
 
   g_object_get(G_OBJECT(connection), "status", &status, NULL);
-  g_return_if_fail(status == INF_XML_CONNECTION_OPENING ||
-                   status == INF_XML_CONNECTION_OPEN);
 
   priv = INF_GTK_BROWSER_STORE_PRIVATE(store);
   item = inf_gtk_browser_store_find_item_by_connection(store, connection);
@@ -2241,4 +2311,130 @@ inf_gtk_browser_store_add_connection(InfGtkBrowserStore* store,
   }
 }
 
+/**
+ * inf_gtk_browser_store_remove_connection:
+ * @store: A #InfGtkBrowserStore.
+ * @connection: A #InfXmlConnection contained in @store.
+ *
+ * This function removes the entry for the given connection from @store.
+ */
+void
+inf_gtk_browser_store_remove_connection(InfGtkBrowserStore* store,
+                                        InfXmlConnection* connection)
+{
+  InfGtkBrowserStoreItem* item;
+
+  g_return_if_fail(INF_GTK_IS_BROWSER_STORE(store));
+  g_return_if_fail(INF_IS_XML_CONNECTION(connection));
+
+  item = inf_gtk_browser_store_find_item_by_connection(store, connection);
+  g_return_if_fail(item != NULL);
+
+  inf_gtk_browser_store_remove_item (store, item);
+}
+
+/**
+ * inf_gtk_browser_store_clear_connection_error:
+ * @store: A #InfGtkBrowserStore.
+ * @connection: A #InfXmlConnection contained in @store.
+ *
+ * This function clears the error for the entry which belongs to the given
+ * connection.
+ */
+void
+inf_gtk_browser_store_clear_connection_error(InfGtkBrowserStore* store,
+                                             InfXmlConnection* connection)
+{
+  InfGtkBrowserStoreItem* item;
+  GtkTreeIter iter;
+  GtkTreePath* path;
+
+  g_return_if_fail(INF_GTK_IS_BROWSER_STORE(store));
+  g_return_if_fail(INF_IS_XML_CONNECTION(connection));
+
+  item = inf_gtk_browser_store_find_item_by_connection(store, connection);
+  g_return_if_fail(item != NULL);
+
+  if(item->status == INF_GTK_BROWSER_MODEL_ERROR)
+  {
+    g_assert(item->error != NULL);
+
+    g_error_free(item->error);
+    item->error = NULL;
+
+    if(item->browser != NULL)
+    {
+      switch(infc_browser_get_status(item->browser))
+      {
+      case INFC_BROWSER_DISCONNECTED:
+        item->status = INF_GTK_BROWSER_MODEL_DISCONNECTED;
+        break;
+      case INFC_BROWSER_CONNECTING:
+        item->status = INF_GTK_BROWSER_MODEL_CONNECTING;
+        break;
+      case INFC_BROWSER_CONNECTED:
+        item->status = INF_GTK_BROWSER_MODEL_CONNECTED;
+        break;
+      default:
+        g_assert_not_reached();
+        break;
+      }
+    }
+    else if(item->info != NULL)
+    {
+      item->status = INF_GTK_BROWSER_MODEL_DISCOVERED;
+    }
+    else
+    {
+      g_assert_not_reached();
+    }
+
+    iter.stamp = INF_GTK_BROWSER_STORE_PRIVATE(store)->stamp;
+    iter.user_data = item;
+    iter.user_data2 = GUINT_TO_POINTER(0);
+    iter.user_data3 = NULL;
+
+    path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
+    gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
+    gtk_tree_path_free (path);
+  }
+}
+
+/**
+ * inf_gtk_browser_store_set_connection_name:
+ * @store: A #InfGtkBrowserStore.
+ * @connection: A #InfXmlConnection contained in @store.
+ * @name: The new name to set for the connection.
+ *
+ * This function sets the name of @connection.
+ **/
+void
+inf_gtk_browser_store_set_connection_name(InfGtkBrowserStore* store,
+                                          InfXmlConnection* connection,
+                                          const gchar* name)
+{
+  InfGtkBrowserStoreItem* item;
+  GtkTreeIter iter;
+  GtkTreePath* path;
+
+  g_return_if_fail(INF_GTK_IS_BROWSER_STORE(store));
+  g_return_if_fail(INF_IS_XML_CONNECTION(connection));
+  g_return_if_fail(name != NULL);
+
+  item = inf_gtk_browser_store_find_item_by_connection(store, connection);
+  g_return_if_fail(item != NULL);
+
+  g_free (item->name);
+  item->name = g_strdup(name);
+
+  iter.stamp = INF_GTK_BROWSER_STORE_PRIVATE(store)->stamp;
+  iter.user_data = item;
+  iter.user_data2 = GUINT_TO_POINTER(0);
+  iter.user_data3 = NULL;
+
+  path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
+  gtk_tree_model_row_changed(GTK_TREE_MODEL(store), path, &iter);
+  gtk_tree_path_free(path);
+}
+
 /* vim:set et sw=2 ts=2: */
