@@ -13,13 +13,15 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfgtk/inf-gtk-browser-view.h>
 #include <libinfinity/common/inf-discovery.h>
 #include <libinfinity/inf-marshal.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <gtk/gtktreeview.h>
 #include <gtk/gtktreeviewcolumn.h>
@@ -120,7 +122,7 @@ static guint view_signals[LAST_SIGNAL];
  * Utility functions
  */
 
-InfGtkBrowserViewBrowser*
+static InfGtkBrowserViewBrowser*
 inf_gtk_browser_view_find_view_browser(InfGtkBrowserView* view,
                                        InfcBrowser* browser)
 {
@@ -139,7 +141,7 @@ inf_gtk_browser_view_find_view_browser(InfGtkBrowserView* view,
   return NULL;
 }
 
-InfGtkBrowserViewExplore*
+static InfGtkBrowserViewExplore*
 inf_gtk_browser_view_find_explore(InfGtkBrowserView* view,
                                   InfGtkBrowserViewBrowser* view_browser,
                                   InfcExploreRequest* request)
@@ -157,7 +159,7 @@ inf_gtk_browser_view_find_explore(InfGtkBrowserView* view,
   return NULL;
 }
 
-InfGtkBrowserViewSync*
+static InfGtkBrowserViewSync*
 inf_gtk_browser_view_find_sync(InfGtkBrowserView* view,
                                InfGtkBrowserViewBrowser* view_browser,
                                InfcSessionProxy* proxy)
@@ -479,19 +481,19 @@ inf_gtk_browser_view_sync_removed(InfGtkBrowserView* view,
     gtk_tree_path_free(path);
   }
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(infc_session_proxy_get_session(sync->proxy)),
     G_CALLBACK(inf_gtk_browser_view_session_synchronization_progress_cb),
     sync
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(infc_session_proxy_get_session(sync->proxy)),
     G_CALLBACK(inf_gtk_browser_view_session_synchronization_complete_cb),
     sync
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(infc_session_proxy_get_session(sync->proxy)),
     G_CALLBACK(inf_gtk_browser_view_session_synchronization_failed_cb),
     sync
@@ -586,19 +588,19 @@ inf_gtk_browser_view_explore_removed(InfGtkBrowserView* view,
     gtk_tree_path_free(path);
   }
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(expl->request),
     G_CALLBACK(inf_gtk_browser_view_explore_request_initiated_cb),
     expl
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(expl->request),
     G_CALLBACK(inf_gtk_browser_view_explore_request_progress_cb),
     expl
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(expl->request),
     G_CALLBACK(inf_gtk_browser_view_explore_request_finished_cb),
     expl
@@ -682,7 +684,8 @@ inf_gtk_browser_view_session_subscribe_cb(InfcBrowser* browser,
    * only be created along with new nodes, in which case we already add the
    * synchronization in row_inserted_cb(). Perhaps we could still
    * double-check here, just to be sure, though... */
-  if(inf_session_get_status(session) == INF_SESSION_SYNCHRONIZING)
+  if(iter != NULL &&
+     inf_session_get_status(session) == INF_SESSION_SYNCHRONIZING)
   {
     result = inf_gtk_browser_model_browser_iter_to_tree_iter(
       INF_GTK_BROWSER_MODEL(model),
@@ -846,48 +849,51 @@ inf_gtk_browser_view_initial_root_explore(InfGtkBrowserView* view,
   view_browser = inf_gtk_browser_view_find_view_browser(view, browser);
   g_assert(view_browser != NULL);
 
-  if(infc_browser_iter_get_explored(browser, browser_iter) == FALSE)
+  if(infc_browser_get_status(browser) == INFC_BROWSER_CONNECTED)
   {
-    request = infc_browser_iter_get_explore_request(browser, browser_iter);
-    /* Explore root node if it is not already explored */
-    if(request == NULL)
-      request = infc_browser_iter_explore(browser, browser_iter);
-
-    if(view_browser->initial_root_expansion == TRUE)
+    if(infc_browser_iter_get_explored(browser, browser_iter) == FALSE)
     {
-      /* There should always only be one view to do initial root expansion
-       * because only one view can have issued the resolv. */
-      g_assert(
-        g_object_get_data(
+      request = infc_browser_iter_get_explore_request(browser, browser_iter);
+      /* Explore root node if it is not already explored */
+      if(request == NULL)
+        request = infc_browser_iter_explore(browser, browser_iter);
+
+      if(view_browser->initial_root_expansion == TRUE)
+      {
+        /* There should always only be one view to do initial root expansion
+         * because only one view can have issued the resolv. */
+        g_assert(
+          g_object_get_data(
+            G_OBJECT(request),
+            INF_GTK_BROWSER_VIEW_INITIAL_EXPANSION
+          ) == NULL
+        );
+
+        /* Remember to do initial root expansion when the node has been
+         * explored. */
+        /*printf("Initial root exansion was set, set on request for view %p\n", view);*/
+        g_object_set_data(
           G_OBJECT(request),
-          INF_GTK_BROWSER_VIEW_INITIAL_EXPANSION
-        ) == NULL
-      );
+          INF_GTK_BROWSER_VIEW_INITIAL_EXPANSION,
+          view
+        );
 
-      /* Remember to do initial root expansion when the node has been
-       * explored. */
-      /*printf("Initial root exansion was set, set on request for view %p\n", view);*/
-      g_object_set_data(
-        G_OBJECT(request),
-        INF_GTK_BROWSER_VIEW_INITIAL_EXPANSION,
-        view
-      );
-
-      /* Handled expansion flag, so unset, could otherwise lead to another
-       * try of expanding the root node. */
-      view_browser->initial_root_expansion = FALSE;
+        /* Handled expansion flag, so unset, could otherwise lead to another
+         * try of expanding the root node. */
+        view_browser->initial_root_expansion = FALSE;
+      }
     }
-  }
-  else
-  {
-    if(view_browser->initial_root_expansion == TRUE)
+    else
     {
-      /*printf("Direct expansion for view %p\n", view);*/
-      gtk_tree_view_expand_row(GTK_TREE_VIEW(priv->treeview), path, FALSE);
+      if(view_browser->initial_root_expansion == TRUE)
+      {
+        /*printf("Direct expansion for view %p\n", view);*/
+        gtk_tree_view_expand_row(GTK_TREE_VIEW(priv->treeview), path, FALSE);
 
-      /* Handled expansion flag, so unset, could otherwise lead to another
-       * try of expanding the root node. */
-      view_browser->initial_root_expansion = FALSE;
+        /* Handled expansion flag, so unset, could otherwise lead to another
+         * try of expanding the root node. */
+        view_browser->initial_root_expansion = FALSE;
+      }
     }
   }
 
@@ -956,6 +962,9 @@ inf_gtk_browser_view_browser_added(InfGtkBrowserView* view,
   if(info != NULL && g_slist_find(priv->info_resolvs, info) != NULL)
   {
     /*printf("Set initial root expansion for view %p\n", view);*/
+    /* TODO: Remember to unset the flag when an error happens or the
+     * corresponding browser is disconnected for another reason before we
+     * actually get to explore anything. */
     view_browser->initial_root_expansion = TRUE;
     priv->info_resolvs = g_slist_remove(priv->info_resolvs, info);
   }
@@ -1000,13 +1009,13 @@ inf_gtk_browser_view_browser_removed(InfGtkBrowserView* view,
   while(view_browser->syncs != NULL)
     inf_gtk_browser_view_sync_removed(view, view_browser->syncs->data);
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(view_browser->browser),
     G_CALLBACK(inf_gtk_browser_view_begin_explore_cb),
     view_browser
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(view_browser->browser),
     G_CALLBACK(inf_gtk_browser_view_session_subscribe_cb),
     view_browser
@@ -1327,37 +1336,37 @@ inf_gtk_browser_view_set_model(InfGtkBrowserView* view,
     g_slist_free(priv->info_resolvs);
     priv->info_resolvs = NULL;
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_row_inserted_cb),
       view
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_row_deleted_cb),
       view
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_row_changed_cb),
       view
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_rows_reordered_cb),
       view
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_set_browser_cb_before),
       view
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(current_model),
       G_CALLBACK(inf_gtk_browser_view_set_browser_cb_after),
       view
@@ -1505,6 +1514,10 @@ inf_gtk_browser_view_row_activated_cb(GtkTreeView* tree_view,
 
   InfcBrowser* browser;
   InfcBrowserIter* browser_iter;
+  InfXmlConnection* connection;
+  InfXmlConnectionStatus xml_status;
+  GError* error;
+  InfGtkBrowserViewBrowser* view_browser;
 
   view = INF_GTK_BROWSER_VIEW(user_data);
   priv = INF_GTK_BROWSER_VIEW_PRIVATE(view);
@@ -1521,14 +1534,44 @@ inf_gtk_browser_view_row_activated_cb(GtkTreeView* tree_view,
       INF_GTK_BROWSER_MODEL_COL_STATUS, &status,
       INF_GTK_BROWSER_MODEL_COL_DISCOVERY, &discovery,
       INF_GTK_BROWSER_MODEL_COL_DISCOVERY_INFO, &info,
+      INF_GTK_BROWSER_MODEL_COL_BROWSER, &browser,
       -1
     );
 
-    if(discovery != NULL)
+    if(browser != NULL)
+    {
+      g_assert(infc_browser_get_connection(browser) != NULL);
+      if(infc_browser_get_status(browser) == INFC_BROWSER_DISCONNECTED)
+      {
+        connection = infc_browser_get_connection(browser);
+        g_assert(connection != NULL);
+        g_object_get(G_OBJECT(connection), "status", &xml_status, NULL);
+        if(xml_status == INF_XML_CONNECTION_CLOSED)
+        {
+          error = NULL;
+          if(!inf_xml_connection_open(connection, &error))
+          {
+            /* TODO: We can't properly report error here. Actually, we should
+             * not do this, but just emit signal activate here, for others
+             * to open the connection if necessary. */
+            g_warning("Failed to reconnect: %s\n", error->message);
+            g_error_free(error);
+          }
+
+          view_browser =
+            inf_gtk_browser_view_find_view_browser(view, browser);
+          g_assert(view_browser != NULL);
+          view_browser->initial_root_expansion = TRUE;
+        }
+      }
+    }
+    else if(discovery != NULL)
     {
       if(status == INF_GTK_BROWSER_MODEL_DISCOVERED ||
          status == INF_GTK_BROWSER_MODEL_ERROR)
       {
+        /* TODO: This method should not exist. Instead, we should just
+         * emit ACTIVATE and make others resolve stuff there. */
         inf_gtk_browser_model_resolve(
           INF_GTK_BROWSER_MODEL(model),
           discovery,
@@ -1540,9 +1583,13 @@ inf_gtk_browser_view_row_activated_cb(GtkTreeView* tree_view,
         priv->info_resolvs = g_slist_prepend(priv->info_resolvs, info);
         /*printf("Add info %p to info resolvs of view %p\n", info, view);*/
       }
-
-      g_object_unref(G_OBJECT(discovery));
     }
+
+    if(discovery != NULL)
+      g_object_unref(G_OBJECT(discovery));
+
+    if(browser != NULL)
+      g_object_unref(browser);
   }
   else
   {
@@ -1881,6 +1928,7 @@ inf_gtk_browser_view_status_icon_data_func(GtkTreeViewColumn* column,
 
     switch(status)
     {
+    case INF_GTK_BROWSER_MODEL_DISCONNECTED:
     case INF_GTK_BROWSER_MODEL_DISCOVERED:
     case INF_GTK_BROWSER_MODEL_RESOLVING:
     case INF_GTK_BROWSER_MODEL_CONNECTING:
@@ -2129,6 +2177,7 @@ inf_gtk_browser_view_status_data_func(GtkTreeViewColumn* column,
     switch(status)
     {
     case INF_GTK_BROWSER_MODEL_DISCOVERED:
+    case INF_GTK_BROWSER_MODEL_DISCONNECTED:
       g_object_set(
         G_OBJECT(renderer),
         "text", _("Not connected"),
