@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfgtk/inf-gtk-certificate-dialog.h>
@@ -626,8 +627,7 @@ inf_gtk_certificate_dialog_set_property(GObject* object,
 
     break;
   case PROP_CERTIFICATE_FLAGS:
-    priv->certificate_flags =
-      (InfGtkCertificateDialogFlags)g_value_get_flags(value);
+    priv->certificate_flags = g_value_get_flags(value);
 
     if(priv->certificate_flags != 0 && priv->hostname != NULL)
       inf_gtk_certificate_dialog_renew_info(dialog);
