@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /* TODO: Put the non-GUI-relevant parts of this code into libinfinity */
@@ -25,6 +26,7 @@
 #include <libinfinity/common/inf-xml-connection.h>
 #include <libinfinity/common/inf-cert-util.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <gnutls/x509.h>
 
@@ -98,7 +100,7 @@ inf_gtk_certificate_manager_notify_status_cb(GObject* object,
 static void
 inf_gtk_certificate_manager_query_free(InfGtkCertificateManagerQuery* query)
 {
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(query->connection),
     G_CALLBACK(inf_gtk_certificate_manager_notify_status_cb),
     query
@@ -136,8 +138,6 @@ inf_gtk_certificate_manager_response_cb(GtkDialog* dialog,
    * accepted the connection, then we will add the new certificate below. */
   if(flags & INF_GTK_CERTIFICATE_DIALOG_CERT_CHANGED)
   {
-    /* TODO: Should we always do this, independant of the
-     * checkbutton setting? */
     if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(query->checkbutton)))
     {
       /* Make sure the certificate is not currently in use by
@@ -452,40 +452,49 @@ inf_gtk_certificate_manager_certificate_func(InfXmppConnection* connection,
     }
 
     own_hostname = inf_cert_util_get_hostname(own);
-    for(i = 0; i < priv->known_hosts->len; ++ i)
+    /* We don't ever trust hosts for which we have a certificate without
+     * hostname, as we can't identify the certificate with the corresponding
+     * host */
+    if(own_hostname != NULL)
     {
-      known = (gnutls_x509_crt_t)g_ptr_array_index(priv->known_hosts, i);
-      if(gnutls_x509_crt_check_hostname(known, own_hostname))
+      for(i = 0; i < priv->known_hosts->len; ++ i)
       {
-        /* TODO: Compare this as binary, not as string */
-        own_fingerprint =
-          inf_cert_util_get_fingerprint(own, GNUTLS_DIG_SHA1);
-        known_fingerprint =
-          inf_cert_util_get_fingerprint(known, GNUTLS_DIG_SHA1);
-
-        if(strcmp(own_fingerprint, known_fingerprint) == 0)
+        known = (gnutls_x509_crt_t)g_ptr_array_index(priv->known_hosts, i);
+        if(gnutls_x509_crt_check_hostname(known, own_hostname))
         {
-          /* We know this host, so we trust it, even if the issuer is 
-           * not a CA. */
-          flags &= ~INF_GTK_CERTIFICATE_DIALOG_CERT_ISSUER_NOT_TRUSTED;
-        }
-        else
-        {
-          /* The fingerprint does not match, so the certificate for this host
-           * has changed. */
-          flags |= INF_GTK_CERTIFICATE_DIALOG_CERT_CHANGED;
+          /* TODO: Compare this as binary, not as string */
+          own_fingerprint =
+            inf_cert_util_get_fingerprint(own, GNUTLS_DIG_SHA1);
+          known_fingerprint =
+            inf_cert_util_get_fingerprint(known, GNUTLS_DIG_SHA1);
 
-          /* Check whether it has changed because the old one expired
-           * (then we have expected the certificate change, otherwise
-           * something strange is going on). */
-          t = gnutls_x509_crt_get_expiration_time(known);
-          if(t == (time_t)(-1) || t < time(NULL))
-            flags |= INF_GTK_CERTIFICATE_DIALOG_CERT_OLD_EXPIRED;
-        }
+          if(strcmp(own_fingerprint, known_fingerprint) == 0)
+          {
+            /* We know this host, so we trust it, even if the issuer is 
+             * not a CA. */
+            flags &= ~INF_GTK_CERTIFICATE_DIALOG_CERT_ISSUER_NOT_TRUSTED;
+          }
+          else
+          {
+            /* The fingerprint does not match, so the certificate for this
+             * host has changed. If the new cert is CA signed we don't care. */
+            if(flags & INF_GTK_CERTIFICATE_DIALOG_CERT_ISSUER_NOT_TRUSTED)
+            {
+              flags |= INF_GTK_CERTIFICATE_DIALOG_CERT_CHANGED;
 
-        g_free(own_fingerprint);
-        g_free(known_fingerprint);
-        break;
+              /* Check whether it has changed because the old one expired
+               * (then we have expected the certificate change, otherwise
+               * something strange is going on). */
+              t = gnutls_x509_crt_get_expiration_time(known);
+              if(t == (time_t)(-1) || t < time(NULL))
+                flags |= INF_GTK_CERTIFICATE_DIALOG_CERT_OLD_EXPIRED;
+            }
+          }
+
+          g_free(own_fingerprint);
+          g_free(known_fingerprint);
+          break;
+        }
       }
     }
 
@@ -578,16 +587,16 @@ inf_gtk_certificate_manager_certificate_func(InfXmppConnection* connection,
       );
 
       text = g_strdup_printf(
-        _("Remember the answer for future connections to host %s"),
+        _("Trust the certificate of host %s in the future"),
         hostname
       );
 
       query->checkbutton = gtk_check_button_new_with_label(text);
 
-      /* TODO: Be able remember any answer, not only the one to
+      /* TODO: Be able to remember any answer, not only the one to
        * "issuer not trusted" */
-      if(flags & INF_GTK_CERTIFICATE_DIALOG_CERT_ISSUER_NOT_TRUSTED ||
-         flags & INF_GTK_CERTIFICATE_DIALOG_CERT_CHANGED)
+      if((flags & INF_GTK_CERTIFICATE_DIALOG_CERT_ISSUER_NOT_TRUSTED) ||
+         (flags & INF_GTK_CERTIFICATE_DIALOG_CERT_CHANGED) )
       {
         gtk_widget_show(query->checkbutton);
       }
