@@ -13,11 +13,13 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfinity/adopted/inf-adopted-algorithm.h>
 #include <libinfinity/inf-marshal.h>
+#include <libinfinity/inf-signals.h>
 
 /**
  * SECTION:inf-adopted-algorithm
@@ -66,6 +68,11 @@
 
 /* TODO: Do only cleanup if too much entries in cache? */
 
+/* TODO: If users are not issuing any requests for some time, and we can be
+ * sure that we do not need to transform any requests, then remove them from
+ * the users array (users_begin, users_end). Readd users as soon as they
+ * issue buffer-altering requests. */
+
 typedef struct _InfAdoptedAlgorithmLocalUser InfAdoptedAlgorithmLocalUser;
 struct _InfAdoptedAlgorithmLocalUser {
   InfAdoptedUser* user;
@@ -888,12 +895,6 @@ inf_adopted_algorithm_can_cache(InfAdoptedRequest* request)
   return flags == INF_ADOPTED_OPERATION_CACHABLE;
 }
 
-/* Required by inf_adopted_algorithm_transform_request() */
-static InfAdoptedRequest*
-inf_adopted_algorithm_translate_request(InfAdoptedAlgorithm* algorithm,
-                                        InfAdoptedRequest* request,
-                                        InfAdoptedStateVector* to);
-
 /* Translates two requests to state at and then transforms them against each
  * other. The result needs to be unref()ed. */
 static InfAdoptedRequest*
@@ -983,7 +984,8 @@ inf_adopted_algorithm_transform_request(InfAdoptedAlgorithm* algorithm,
 }
 
 /* This should only ever be called by
- * inf_adopted_algorithm_translate_request() after it did the cache checks */
+ * inf_adopted_algorithm_translate_request() after it did the cache and
+ * validity checks */
 static InfAdoptedRequest*
 inf_adopted_algorithm_translate_request_nocache(InfAdoptedAlgorithm* algorithm,
                                                 InfAdoptedRequest* request,
@@ -999,34 +1001,19 @@ inf_adopted_algorithm_translate_request_nocache(InfAdoptedAlgorithm* algorithm,
   InfAdoptedRequest* associated;
   InfAdoptedRequest* translated;
   InfAdoptedRequest* result;
-  /*InfAdoptedStateVector* v;*/
   gint by;
   guint n;
 
   priv = INF_ADOPTED_ALGORITHM_PRIVATE(algorithm);
-  user_id = inf_adopted_request_get_user_id(request);
-  user = inf_user_table_lookup_user_by_id(priv->user_table, user_id);
-  g_assert(user != NULL);
-
-  log = inf_adopted_user_get_request_log(INF_ADOPTED_USER(user));
-
-  g_assert(inf_adopted_state_vector_causally_before(to, priv->current));
-  g_assert(
-    inf_adopted_state_vector_causally_before(
-      inf_adopted_request_get_vector(
-        inf_adopted_request_log_original_request(log, request)
-      ),
-      to
-    )
-  );
-  g_assert(inf_adopted_algorithm_is_reachable(algorithm, to) == TRUE);
-
   vector = inf_adopted_request_get_vector(request);
-  /*v = inf_adopted_state_vector_copy(to);*/
   result = NULL;
 
   if(inf_adopted_request_get_request_type(request) != INF_ADOPTED_REQUEST_DO)
   {
+    user_id = inf_adopted_request_get_user_id(request);
+    user = inf_user_table_lookup_user_by_id(priv->user_table, user_id);
+    log = inf_adopted_user_get_request_log(INF_ADOPTED_USER(user));
+
     /* Try late mirror if this is not a do request */
     associated = inf_adopted_request_log_prev_associated(log, request);
     g_assert(associated != NULL);
@@ -1184,63 +1171,6 @@ inf_adopted_algorithm_translate_request_nocache(InfAdoptedAlgorithm* algorithm,
   return NULL;
 }
 
-/* The result needs to be unref()ed. */
-static InfAdoptedRequest*
-inf_adopted_algorithm_translate_request(InfAdoptedAlgorithm* algorithm,
-                                        InfAdoptedRequest* request,
-                                        InfAdoptedStateVector* to)
-{
-  InfAdoptedAlgorithmPrivate* priv;
-  guint user_id;
-
-  InfAdoptedRequest* result;
-
-  InfAdoptedAlgorithmRequestKey lookup_key;
-  InfAdoptedAlgorithmRequestKey* insert_key;
-
-  priv = INF_ADOPTED_ALGORITHM_PRIVATE(algorithm);
-  user_id = inf_adopted_request_get_user_id(request);
-
-  /* If the request affects the buffer, then it might have been cached
-   * earlier. */
-  if(inf_adopted_request_affects_buffer(request))
-  {
-    lookup_key.vector = to;
-    lookup_key.user_id = user_id;
-    result = g_tree_lookup(priv->cache, &lookup_key);
-    if(result != NULL)
-    {
-      g_object_ref(result);
-      return result;
-    }
-  }
-
-  result = inf_adopted_algorithm_translate_request_nocache(
-    algorithm,
-    request,
-    to
-  );
-
-  g_assert(
-    inf_adopted_state_vector_compare(
-      inf_adopted_request_get_vector(result),
-      to
-    ) == 0
-  );
-
-  if(inf_adopted_algorithm_can_cache(result))
-  {
-    insert_key = g_slice_new(InfAdoptedAlgorithmRequestKey);
-    insert_key->vector = inf_adopted_request_get_vector(result);
-    insert_key->user_id = user_id;
-    g_assert(g_tree_lookup(priv->cache, insert_key) == NULL);
-    g_tree_replace(priv->cache, insert_key, result);
-    g_object_ref(result);
-  }
-
-  return result;
-}
-
 static void
 inf_adopted_algorithm_init(GTypeInstance* instance,
                            gpointer g_class)
@@ -1363,7 +1293,7 @@ inf_adopted_algorithm_dispose(GObject* object)
 
   if(priv->buffer != NULL)
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->buffer),
       G_CALLBACK(inf_adopted_algorithm_buffer_notify_modified_cb),
       algorithm
@@ -1378,19 +1308,19 @@ inf_adopted_algorithm_dispose(GObject* object)
 
   if(priv->user_table != NULL)
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->user_table),
       G_CALLBACK(inf_adopted_algorithm_add_user_cb),
       algorithm
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->user_table),
       G_CALLBACK(inf_adopted_algorithm_add_local_user_cb),
       algorithm
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->user_table),
       G_CALLBACK(inf_adopted_algorithm_remove_local_user_cb),
       algorithm
@@ -1682,7 +1612,7 @@ inf_adopted_algorithm_execute_request(InfAdoptedAlgorithm* algorithm,
     g_object_unref(log_request);
   }
 
-  g_signal_handlers_block_by_func(
+  inf_signal_handlers_block_by_func(
     G_OBJECT(priv->buffer),
     G_CALLBACK(inf_adopted_algorithm_buffer_notify_modified_cb),
     algorithm
@@ -1730,7 +1660,7 @@ inf_adopted_algorithm_execute_request(InfAdoptedAlgorithm* algorithm,
     g_assert(inf_buffer_get_modified(priv->buffer) == TRUE);
   }
 
-  g_signal_handlers_unblock_by_func(
+  inf_signal_handlers_unblock_by_func(
     G_OBJECT(priv->buffer),
     G_CALLBACK(inf_adopted_algorithm_buffer_notify_modified_cb),
     algorithm
@@ -1842,7 +1772,7 @@ inf_adopted_algorithm_class_init(gpointer g_class,
 
   /**
    * InfAdoptedAlgorithm::can-undo-changed:
-   * @algorithm: The #InfAdopedAlgorithm for which a user's
+   * @algorithm: The #InfAdoptedAlgorithm for which a user's
    * can-undo state changed.
    * @user: The #InfAdoptedUser whose can-undo state has changed.
    * @can_undo: Whether @user can issue an undo request in the current
@@ -1869,7 +1799,7 @@ inf_adopted_algorithm_class_init(gpointer g_class,
 
   /**
    * InfAdoptedAlgorithm::can-redo-changed:
-   * @algorithm: The #InfAdopedAlgorithm for which a user's
+   * @algorithm: The #InfAdoptedAlgorithm for which a user's
    * can-redo state changed.
    * @user: The #InfAdoptedUser whose can-redo state has changed.
    * @can_undo: Whether @user can issue a redo request in the current
@@ -2279,6 +2209,117 @@ inf_adopted_algorithm_generate_redo(InfAdoptedAlgorithm* algorithm,
   return request;
 }
 
+/**
+ * inf_adopted_algorithm_translate_request:
+ * @algorithm: A #InfAdoptedAlgorithm.
+ * @request: A #InfAdoptedRequest.
+ * @to: The state vector to translate @request to.
+ *
+ * Translates @request so that it can be applied to the document at state @to.
+ * @request will not be modified but a new, translated request is returned
+ * instead.
+ *
+ * There are several preconditions for this function to be called. @to must
+ * be a reachable point in the state space. Also, requests can only be
+ * translated in forward direction, so @request's vector time must be
+ * causally before (see inf_adopted_state_vector_causally_before()) @to.
+ *
+ * Returns: A new or cached #InfAdoptedRequest. Free with g_object_unref()
+ * when no longer needed.
+ */
+InfAdoptedRequest*
+inf_adopted_algorithm_translate_request(InfAdoptedAlgorithm* algorithm,
+                                        InfAdoptedRequest* request,
+                                        InfAdoptedStateVector* to)
+{
+  InfAdoptedAlgorithmPrivate* priv;
+  guint user_id;
+
+  InfAdoptedRequest* result;
+
+  InfAdoptedAlgorithmRequestKey lookup_key;
+  InfAdoptedAlgorithmRequestKey* insert_key;
+
+  g_return_val_if_fail(INF_ADOPTED_IS_ALGORITHM(algorithm), NULL);
+  g_return_val_if_fail(INF_ADOPTED_IS_REQUEST(request), NULL);
+  g_return_val_if_fail(to != NULL, NULL);
+
+  priv = INF_ADOPTED_ALGORITHM_PRIVATE(algorithm);
+  user_id = inf_adopted_request_get_user_id(request);
+
+  /* Validity checks */
+  g_return_val_if_fail(
+    inf_user_table_lookup_user_by_id(priv->user_table, user_id) != NULL,
+    NULL
+  );
+
+  g_return_val_if_fail(
+    inf_adopted_state_vector_causally_before(to, priv->current),
+    NULL
+  );
+
+  g_return_val_if_fail(
+    inf_adopted_state_vector_causally_before(
+      inf_adopted_request_get_vector(
+        inf_adopted_request_log_original_request(
+          inf_adopted_user_get_request_log(
+            INF_ADOPTED_USER(
+              inf_user_table_lookup_user_by_id(priv->user_table, user_id)
+            )
+          ),
+          request
+        )
+      ),
+      to
+    ),
+    NULL
+  );
+
+  g_return_val_if_fail(
+    inf_adopted_algorithm_is_reachable(algorithm, to) == TRUE,
+    NULL
+  );
+
+  /* If the request affects the buffer, then it might have been cached
+   * earlier. */
+  if(inf_adopted_request_affects_buffer(request))
+  {
+    lookup_key.vector = to;
+    lookup_key.user_id = user_id;
+    result = g_tree_lookup(priv->cache, &lookup_key);
+    if(result != NULL)
+    {
+      g_object_ref(result);
+      return result;
+    }
+  }
+
+  result = inf_adopted_algorithm_translate_request_nocache(
+    algorithm,
+    request,
+    to
+  );
+
+  g_assert(
+    inf_adopted_state_vector_compare(
+      inf_adopted_request_get_vector(result),
+      to
+    ) == 0
+  );
+
+  if(inf_adopted_algorithm_can_cache(result))
+  {
+    insert_key = g_slice_new(InfAdoptedAlgorithmRequestKey);
+    insert_key->vector = inf_adopted_request_get_vector(result);
+    insert_key->user_id = user_id;
+    g_assert(g_tree_lookup(priv->cache, insert_key) == NULL);
+    g_tree_replace(priv->cache, insert_key, result);
+    g_object_ref(result);
+  }
+
+  return result;
+}
+
 /**
  * inf_adopted_algorithm_receive_request:
  * @algorithm: A #InfAdoptedAlgorithm.
