@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /* TODO: warning if no update from a particular non-local user for some time */
@@ -21,7 +22,9 @@
 #include <libinfinity/adopted/inf-adopted-session.h>
 #include <libinfinity/adopted/inf-adopted-no-operation.h>
 #include <libinfinity/common/inf-xml-util.h>
+#include <libinfinity/common/inf-error.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <string.h>
 #include <time.h>
@@ -94,7 +97,7 @@ static const int INF_ADOPTED_SESSION_NOOP_INTERVAL = 30;
  * Utility functions.
  */
 
-InfAdoptedSessionLocalUser*
+static InfAdoptedSessionLocalUser*
 inf_adopted_session_lookup_local_user(InfAdoptedSession* session,
                                       InfAdoptedUser* user)
 {
@@ -163,6 +166,7 @@ inf_adopted_session_validate_request(InfAdoptedRequestLog* log,
           error,
           inf_adopted_session_error_quark,
           INF_ADOPTED_SESSION_ERROR_INVALID_REQUEST,
+          "%s",
           _("Undo received, but no previous request found")
         );
 
@@ -179,6 +183,7 @@ inf_adopted_session_validate_request(InfAdoptedRequestLog* log,
           error,
           inf_adopted_session_error_quark,
           INF_ADOPTED_SESSION_ERROR_INVALID_REQUEST,
+          "%s",
           _("Redo received, but no previous request found")
         );
 
@@ -361,6 +366,62 @@ inf_adopted_session_stop_noop_timer(InfAdoptedSession* session,
   }
 }
 
+/* Breadcasts a request N times - makes only sense for undo and redo requests,
+ * so that's the only thing we offer API for. */
+static void
+inf_adopted_session_broadcast_n_requests(InfAdoptedSession* session,
+                                         InfAdoptedRequest* request,
+                                         guint n)
+{
+  InfAdoptedSessionPrivate* priv;
+  InfAdoptedSessionClass* session_class;
+  InfUserTable* user_table;
+  guint user_id;
+  InfUser* user;
+  InfAdoptedSessionLocalUser* local;
+  xmlNodePtr xml;
+
+  priv = INF_ADOPTED_SESSION_PRIVATE(session);
+  session_class = INF_ADOPTED_SESSION_GET_CLASS(session);
+  g_assert(session_class->request_to_xml != NULL);
+
+  user_table = inf_session_get_user_table(INF_SESSION(session));
+  user_id = inf_adopted_request_get_user_id(request);
+  user = inf_user_table_lookup_user_by_id(user_table, user_id);
+  g_assert(user != NULL);
+
+  local = inf_adopted_session_lookup_local_user(
+    session,
+    INF_ADOPTED_USER(user)
+  );
+  g_assert(local != NULL);
+
+  xml = xmlNewNode(NULL, (const xmlChar*)"request");
+
+  session_class->request_to_xml(
+    session,
+    xml,
+    request,
+    local->last_send_vector,
+    FALSE
+  );
+
+  if(n > 1) inf_xml_util_set_attribute_uint(xml, "num", n);
+  inf_session_send_to_subscriptions(INF_SESSION(session), xml);
+
+  inf_adopted_state_vector_free(local->last_send_vector);
+  local->last_send_vector = inf_adopted_state_vector_copy(
+    inf_adopted_request_get_vector(request)
+  );
+
+  /* Add this request to last send vector if it increases vector time
+   * (-> affects buffer). */
+  if(inf_adopted_request_affects_buffer(request) == TRUE)
+    inf_adopted_state_vector_add(local->last_send_vector, user_id, n);
+
+  inf_adopted_session_stop_noop_timer(session, local);
+}
+
 /*
  * Signal handlers
  */
@@ -494,6 +555,9 @@ inf_adopted_session_execute_request_cb(InfAdoptedAlgorithm* algorithm,
   if(inf_adopted_request_get_request_type(request) != INF_ADOPTED_REQUEST_DO ||
      !INF_ADOPTED_IS_NO_OPERATION(inf_adopted_request_get_operation(request)))
   {
+    /* TODO: We should offer a virtual function to flush all requests for
+     * local users, either here or even in InfSession via a vfunc, so that
+     * we don't accidentally make local users active by a delayed request. */
     if(inf_user_get_status(INF_USER(user)) == INF_USER_INACTIVE)
       g_object_set(G_OBJECT(user), "status", INF_USER_ACTIVE, NULL);
   }
@@ -594,6 +658,7 @@ inf_adopted_session_constructor(GType type,
 
   switch(status)
   {
+  case INF_SESSION_PRESYNC:
   case INF_SESSION_SYNCHRONIZING:
     /* algorithm is created during initial synchronization when parameters
      * like initial vector time, max total log size etc. are known. */
@@ -633,13 +698,13 @@ inf_adopted_session_dispose(GObject* object)
 
   user_table = inf_session_get_user_table(INF_SESSION(session));
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(user_table),
     G_CALLBACK(inf_adopted_session_add_local_user_cb),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(user_table),
     G_CALLBACK(inf_adopted_session_remove_local_user_cb),
     session
@@ -659,7 +724,7 @@ inf_adopted_session_dispose(GObject* object)
 
   if(priv->algorithm != NULL)
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->algorithm),
       G_CALLBACK(inf_adopted_session_execute_request_cb),
       session
@@ -715,7 +780,7 @@ inf_adopted_session_set_property(GObject* object,
   case PROP_ALGORITHM:
     /* read only */
   default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID(value, prop_id, pspec);
+    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
     break;
   }
 }
@@ -877,6 +942,12 @@ inf_adopted_session_process_xml_run(InfSession* session,
   InfAdoptedRequest* request;
   InfAdoptedUser* user;
 
+  gboolean has_num;
+  guint num;
+  GError* local_error;
+  InfAdoptedRequest* copy_req;
+  guint i;
+
   priv = INF_ADOPTED_SESSION_PRIVATE(session);
 
   if(strcmp((const char*)xml->name, "request") == 0)
@@ -893,8 +964,28 @@ inf_adopted_session_process_xml_run(InfSession* session,
     if(user == NULL)
       return INF_COMMUNICATION_SCOPE_PTP;
 
-    /* TODO: Check user connection! */
+    if(inf_user_get_status(INF_USER(user)) == INF_USER_UNAVAILABLE ||
+       inf_user_get_connection(INF_USER(user)) != connection)
+    {
+      g_set_error(
+        error,
+        inf_user_error_quark(),
+        INF_USER_ERROR_NOT_JOINED,
+        "%s",
+        _("User did not join from this connection")
+      );
 
+      return INF_COMMUNICATION_SCOPE_PTP;
+    }
+
+    local_error = NULL;
+    has_num = inf_xml_util_get_attribute_uint(xml, "num", &num, &local_error);
+    if(local_error != NULL)
+    {
+      g_propagate_error(error, local_error);
+      return INF_COMMUNICATION_SCOPE_PTP;
+    }
+    
     request = session_class->xml_to_request(
       INF_ADOPTED_SESSION(session),
       xml,
@@ -907,7 +998,30 @@ inf_adopted_session_process_xml_run(InfSession* session,
       return INF_COMMUNICATION_SCOPE_PTP;
 
     inf_adopted_algorithm_receive_request(priv->algorithm, request);
-    g_object_unref(G_OBJECT(request));
+
+    /* Apply the request more than once if num is given. This is mostly used
+     * for multiple undos and redos, but is in general allowed for any
+     * request. */
+    if(has_num)
+    {
+      for(i = 1; i < num; ++i)
+      {
+        /* TODO: This is a bit of a hack since requests are normally
+         * immutable. It avoids an additional vector copy here though. */
+        copy_req = inf_adopted_request_copy(request);
+
+        inf_adopted_state_vector_add(
+          inf_adopted_request_get_vector(copy_req),
+          inf_user_get_id(INF_USER(user)),
+          i
+        );
+
+        inf_adopted_algorithm_receive_request(priv->algorithm, copy_req);
+        g_object_unref(copy_req);
+      }
+    }
+
+    g_object_unref(request);
 
     /* Requests can always be forwarded since user is given. */
     return INF_COMMUNICATION_SCOPE_GROUP;
@@ -1046,6 +1160,7 @@ inf_adopted_session_validate_user_props(InfSession* session,
       error,
       inf_adopted_session_error_quark,
       INF_ADOPTED_SESSION_ERROR_MISSING_STATE_VECTOR,
+      "%s",
       _("'time' attribute in user message is missing")
     );
 
@@ -1268,8 +1383,8 @@ inf_adopted_session_get_io(InfAdoptedSession* session)
  * @session: A #InfAdoptedSession.
  *
  * Returns the #InfAlgorithm object of @session. Returns %NULL if @session
- * has status %INF_SESSION_SYNCHRONIZING because there the algorithm object
- * is not yet created before successful synchronization.
+ * has status %INF_SESSION_PRESYNC or %INF_SESSION_SYNCHRONIZING because there
+ * the algorithm object is not yet created before successful synchronization.
  *
  * Return Value: A #InfAdoptedAlgorithm, or %NULL.
  **/
@@ -1295,77 +1410,41 @@ void
 inf_adopted_session_broadcast_request(InfAdoptedSession* session,
                                       InfAdoptedRequest* request)
 {
-  InfAdoptedSessionPrivate* priv;
-  InfAdoptedSessionClass* session_class;
-  InfUserTable* user_table;
-  guint user_id;
-  InfUser* user;
-  InfAdoptedSessionLocalUser* local;
-  xmlNodePtr xml;
-
   g_return_if_fail(INF_ADOPTED_IS_SESSION(session));
   g_return_if_fail(INF_ADOPTED_IS_REQUEST(request));
 
-  priv = INF_ADOPTED_SESSION_PRIVATE(session);
-  session_class = INF_ADOPTED_SESSION_GET_CLASS(session);
-  g_assert(session_class->request_to_xml != NULL);
-
-  user_table = inf_session_get_user_table(INF_SESSION(session));
-  user_id = inf_adopted_request_get_user_id(request);
-  user = inf_user_table_lookup_user_by_id(user_table, user_id);
-  g_return_if_fail(user != NULL);
-
-  local = inf_adopted_session_lookup_local_user(
-    session,
-    INF_ADOPTED_USER(user)
-  );
-  g_return_if_fail(local != NULL);
-
-  xml = xmlNewNode(NULL, (const xmlChar*)"request");
-
-  session_class->request_to_xml(
-    session,
-    xml,
-    request,
-    local->last_send_vector,
-    FALSE
-  );
-
-  inf_session_send_to_subscriptions(INF_SESSION(session), NULL, xml);
-
-  inf_adopted_state_vector_free(local->last_send_vector);
-  local->last_send_vector = inf_adopted_state_vector_copy(
-    inf_adopted_request_get_vector(request)
-  );
-
-  /* Add this request to last send vector if it increases vector time
-   * (-> affects buffer). */
-  if(inf_adopted_request_affects_buffer(request) == TRUE)
-    inf_adopted_state_vector_add(local->last_send_vector, user_id, 1);
-
-  inf_adopted_session_stop_noop_timer(session, local);
+  inf_adopted_session_broadcast_n_requests(session, request, 1);
 }
 
 /**
  * inf_adopted_session_undo:
  * @session: A #InfAdoptedSession.
  * @user: A local #InfAdoptedUser.
+ * @n: The number of undo requests to issue.
  *
- * This is a shortcut for creating an undo request and broadcasting it.
+ * This is a shortcut for creating @n undo requests and broadcasting them.
+ * If @n > 1 then this is also more efficient.
  **/
 void
 inf_adopted_session_undo(InfAdoptedSession* session,
-                         InfAdoptedUser* user)
+                         InfAdoptedUser* user,
+                         guint n)
 {
   InfAdoptedSessionPrivate* priv;
   InfAdoptedRequest* request;
+  guint i;
 
   g_return_if_fail(INF_ADOPTED_IS_SESSION(session));
   g_return_if_fail(INF_ADOPTED_IS_USER(user));
+  g_return_if_fail(n >= 1);
+
+  /* TODO: Check whether we can issue n undo requests before doing anything */
 
   priv = INF_ADOPTED_SESSION_PRIVATE(session);
   request = inf_adopted_algorithm_generate_undo(priv->algorithm, user);
-  inf_adopted_session_broadcast_request(session, request);
+  for(i = 1; i < n; ++i)
+    inf_adopted_algorithm_generate_undo(priv->algorithm, user);
+  inf_adopted_session_broadcast_n_requests(session, request, n);
   g_object_unref(request);  
 }
 
@@ -1373,22 +1452,31 @@ inf_adopted_session_undo(InfAdoptedSession* session,
  * inf_adopted_session_redo:
  * @session: A #InfAdoptedSession.
  * @user: A local #InfAdoptedUser.
+ * @n: The number of redo requests to issue.
  *
- * This is a shortcut for creating a redo request and broadcasting it.
+ * This is a shortcut for creating @n redo requests and broadcasting them.
+ * If @n > 1 then this is also more efficient.
  **/
 void
 inf_adopted_session_redo(InfAdoptedSession* session,
-                         InfAdoptedUser* user)
+                         InfAdoptedUser* user,
+                         guint n)
 {
   InfAdoptedSessionPrivate* priv;
   InfAdoptedRequest* request;
+  guint i;
+
+  /* TODO: Check whether we can issue n redo requests before doing anything */
 
   g_return_if_fail(INF_ADOPTED_IS_SESSION(session));
   g_return_if_fail(INF_ADOPTED_IS_USER(user));
+  g_return_if_fail(n >= 1);
 
   priv = INF_ADOPTED_SESSION_PRIVATE(session);
   request = inf_adopted_algorithm_generate_redo(priv->algorithm, user);
-  inf_adopted_session_broadcast_request(session, request);
+  for(i = 1; i < n; ++i)
+    inf_adopted_algorithm_generate_redo(priv->algorithm, user);
+  inf_adopted_session_broadcast_n_requests(session, request, n);
   g_object_unref(request);
 }
 
@@ -1463,6 +1551,7 @@ inf_adopted_session_read_request_info(InfAdoptedSession* session,
         error,
         inf_adopted_session_error_quark,
         INF_ADOPTED_SESSION_ERROR_MISSING_OPERATION,
+        "%s",
         _("Operation for request missing")
       );
 
