@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /**
@@ -722,6 +723,7 @@ inf_adopted_state_vector_from_string(const gchar* str,
 {
   InfAdoptedStateVector* vec;
   const char* strpos;
+  char* endpos;
   gsize pos;
   guint id;
   guint n;
@@ -733,13 +735,14 @@ inf_adopted_state_vector_from_string(const gchar* str,
 
   while(*strpos)
   {
-    id = strtoul(strpos, (char**)&strpos, 10);
-    if(*strpos != ':')
+    id = strtoul(strpos, &endpos, 10);
+    if(*endpos != ':')
     {
       g_set_error(
         error,
         inf_adopted_state_vector_error_quark(),
         INF_ADOPTED_STATE_VECTOR_BAD_FORMAT,
+        "%s",
         _("Expected ':' after ID")
       );
 
@@ -762,10 +765,10 @@ inf_adopted_state_vector_from_string(const gchar* str,
       return NULL;
     }
 
-    ++ strpos; /* step over ':' */
-    n = strtoul(strpos, (char**)&strpos, 10);
+    strpos = endpos + 1; /* step over ':' */
+    n = strtoul(strpos, &endpos, 10);
 
-    if(*strpos != ';' && *strpos != '\0')
+    if(*endpos != ';' && *endpos != '\0')
     {
       g_set_error(
         error,
@@ -780,6 +783,7 @@ inf_adopted_state_vector_from_string(const gchar* str,
     }
 
     inf_adopted_state_vector_insert(vec, id, n, pos);
+    strpos = endpos;
     if(*strpos != '\0') ++ strpos; /* step over ';' */
   }
 
@@ -886,7 +890,8 @@ inf_adopted_state_vector_to_string_diff(InfAdoptedStateVector* vec,
  * Recreates a vector from its string representation diff and the original
  * vector. If an error returns, the function returns %NULL and @error is set.
  *
- * Return Value:
+ * Return Value: The created state vector, or %NULL on error. Free with
+ * inf_adopted_state_vector_free() when no longer needed.
  **/
 InfAdoptedStateVector*
 inf_adopted_state_vector_from_string_diff(const gchar* str,
