@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfinity/client/infc-session-proxy.h>
@@ -24,6 +25,7 @@
 #include <libinfinity/common/inf-xml-util.h>
 #include <libinfinity/common/inf-error.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <string.h>
 
@@ -44,6 +46,7 @@ enum {
 
   PROP_SESSION,
   PROP_SUBSCRIPTION_GROUP,
+  PROP_SEQUENCE_ID,
   PROP_CONNECTION
 };
 
@@ -54,7 +57,8 @@ static void
 infc_session_proxy_unsubscribe_connection(InfcSessionProxy* proxy)
 {
   InfcSessionProxyPrivate* priv;
-  InfXmlConnectionStatus status;
+  InfSessionSyncStatus sync_status;
+  InfXmlConnectionStatus conn_status;
   xmlNodePtr xml;
 
   priv = INFC_SESSION_PROXY_PRIVATE(proxy);
@@ -62,11 +66,13 @@ infc_session_proxy_unsubscribe_connection(InfcSessionProxy* proxy)
   g_assert(priv->connection != NULL);
   g_assert(priv->session != NULL);
 
-  status = inf_session_get_synchronization_status(
+  sync_status = inf_session_get_synchronization_status(
     priv->session,
     priv->connection
   );
 
+  g_object_get(G_OBJECT(priv->connection), "status", &conn_status, NULL);
+
   /* If synchronization is still in progress, the close default signal
    * handler in InfSession the base class will cancel the synchronization in
    * which case we do not need to send an extra session-unsubscribe
@@ -75,7 +81,13 @@ infc_session_proxy_unsubscribe_connection(InfcSessionProxy* proxy)
   /* However, in case we are in AWAITING_ACK status we send session
    * unsubscribe because we cannot cancel the synchronization anymore but
    * the server will go into RUNNING state before receiving this message. */
-  if(status != INF_SESSION_SYNC_IN_PROGRESS)
+
+  /* Note that, also, the connection might not be open anymore at this point
+   * if a handler for the connection's notify::status signal finalized or
+   * closed the session explicitely before our handler was called. Therefore
+   * we also check the connection status here. */
+  if(conn_status == INF_XML_CONNECTION_OPEN &&
+     sync_status != INF_SESSION_SYNC_IN_PROGRESS)
   {
     xml = xmlNewNode(NULL, (const xmlChar*)"session-unsubscribe");
 
@@ -166,7 +178,7 @@ infc_session_proxy_session_synchronization_failed_cb(InfSession* session,
   InfcSessionProxy* proxy;
   InfcSessionProxyPrivate* priv;
   InfSessionStatus status;
-  
+
   proxy = INFC_SESSION_PROXY(user_data);
   priv = INFC_SESSION_PROXY_PRIVATE(proxy);
 
@@ -212,19 +224,19 @@ infc_session_proxy_session_close_cb(InfSession* session,
 
   /* Don't release session so others can still access */
 #if 0
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(priv->session),
     G_CALLBACK(infc_session_proxy_session_close_cb),
     proxy
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(priv->session),
     G_CALLBACK(infc_session_proxy_session_synchronization_complete_cb),
     proxy
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(priv->session),
     G_CALLBACK(infc_session_proxy_session_synchronization_failed_cb),
     proxy
@@ -258,6 +270,9 @@ infc_session_proxy_release_connection(InfcSessionProxy* proxy)
   /* TODO: Emit failed signal with some "cancelled" error? */
   infc_request_manager_clear(priv->request_manager);
 
+  g_object_unref(priv->request_manager);
+  priv->request_manager = NULL;
+
   /* Set status of all users to unavailable */
   /* TODO: Keep local users available */
   inf_user_table_foreach_user(
@@ -266,7 +281,7 @@ infc_session_proxy_release_connection(InfcSessionProxy* proxy)
     NULL
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(priv->connection),
     G_CALLBACK(infc_session_proxy_connection_notify_status_cb),
     proxy
@@ -297,39 +312,6 @@ infc_session_proxy_request_to_xml(InfcRequest* request)
   return xml;
 }
 
-static void
-infc_session_proxy_succeed_user_request(InfcSessionProxy* proxy,
-                                        xmlNodePtr xml,
-                                        InfUser* user)
-{
-  InfcSessionProxyPrivate* priv;
-  xmlChar* seq_attr;
-  guint seq;
-  InfcRequest* request;
-
-  priv = INFC_SESSION_PROXY_PRIVATE(proxy);
-  seq_attr = xmlGetProp(xml, (const xmlChar*)"seq");
-
-  if(seq_attr != NULL)
-  {
-    seq = strtoul((const char*)seq_attr, NULL, 0);
-    xmlFree(seq_attr);
-
-    request = infc_request_manager_get_request_by_seq(
-      priv->request_manager,
-      seq
-    );
-
-    /* TODO: Set error when invalid seq was set. Perhaps implement in
-     * request manager. */
-    if(INFC_IS_USER_REQUEST(request))
-    {
-      infc_user_request_finished(INFC_USER_REQUEST(request), user);
-      infc_request_manager_remove_request(priv->request_manager, request);
-    }
-  }
-}
-
 /*
  * GObject overrides.
  */
@@ -347,7 +329,7 @@ infc_session_proxy_init(GTypeInstance* instance,
   priv->session = NULL;
   priv->subscription_group = NULL;
   priv->connection = NULL;
-  priv->request_manager = infc_request_manager_new();
+  priv->request_manager = NULL;
 }
 
 static void
@@ -365,19 +347,19 @@ infc_session_proxy_dispose(GObject* object)
   /* Release session */
   if(priv->session != NULL)
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->session),
       G_CALLBACK(infc_session_proxy_session_close_cb),
       proxy
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->session),
       G_CALLBACK(infc_session_proxy_session_synchronization_complete_cb),
       proxy
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->session),
       G_CALLBACK(infc_session_proxy_session_synchronization_failed_cb),
       proxy
@@ -387,9 +369,7 @@ infc_session_proxy_dispose(GObject* object)
     priv->session = NULL;
   }
 
-  g_object_unref(G_OBJECT(priv->request_manager));
-  priv->request_manager = NULL;
-
+  g_assert(priv->request_manager == NULL);
   G_OBJECT_CLASS(parent_class)->dispose(object);
 }
 
@@ -521,6 +501,7 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
 {
   InfcSessionProxyPrivate* priv;
   InfSessionClass* session_class;
+  InfcRequest* request;
   GArray* array;
   InfUser* user;
   GParameter* param;
@@ -529,6 +510,13 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
   priv = INFC_SESSION_PROXY_PRIVATE(proxy);
   session_class = INF_SESSION_GET_CLASS(priv->session);
 
+  request = infc_request_manager_get_request_by_xml(
+    priv->request_manager,
+    "user-join",
+    xml,
+    NULL
+  );
+
   array = session_class->get_xml_user_props(priv->session, connection, xml);
 
   /* Set local flag if the join was requested by us (seq is present in
@@ -537,7 +525,7 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
   g_assert(!G_IS_VALUE(&param->value)); /* must not have been set already */
 
   g_value_init(&param->value, INF_TYPE_USER_FLAGS);
-  if(xmlHasProp(xml, (const xmlChar*)"seq") != NULL)
+  if(request != NULL)
     g_value_set_flags(&param->value, INF_USER_LOCAL);
   else
     g_value_set_flags(&param->value, 0);
@@ -564,7 +552,12 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
 
   if(user != NULL)
   {
-    infc_session_proxy_succeed_user_request(proxy, xml, user);
+    if(request != NULL)
+    {
+      infc_user_request_finished(INFC_USER_REQUEST(request), user);
+      infc_request_manager_remove_request(priv->request_manager, request);
+    }
+
     return TRUE;
   }
   else
@@ -581,6 +574,7 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
 {
   InfcSessionProxyPrivate* priv;
   InfSessionClass* session_class;
+  InfcRequest* request;
   GArray* array;
   InfUser* user;
   const GParameter* idparam;
@@ -592,6 +586,13 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
   priv = INFC_SESSION_PROXY_PRIVATE(proxy);
   session_class = INF_SESSION_GET_CLASS(priv->session);
 
+  request = infc_request_manager_get_request_by_xml(
+    priv->request_manager,
+    "user-join",
+    xml,
+    NULL
+  );
+
   array = session_class->get_xml_user_props(priv->session, connection, xml);
 
   /* Find rejoining user first */
@@ -607,6 +608,7 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
       error,
       inf_request_error_quark(),
       INF_REQUEST_ERROR_NO_SUCH_ATTRIBUTE,
+      "%s",
       _("Request does not contain required attribute 'id'")
     );
 
@@ -638,7 +640,7 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
   g_assert(!G_IS_VALUE(&param->value)); /* must not have been set already */
 
   g_value_init(&param->value, INF_TYPE_USER_FLAGS);
-  if(xmlHasProp(xml, (const xmlChar*)"seq") != NULL)
+  if(request != NULL)
     g_value_set_flags(&param->value, INF_USER_LOCAL);
   else
     g_value_set_flags(&param->value, 0);
@@ -684,7 +686,11 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
     g_value_unset(&g_array_index(array, GParameter, i).value);
   g_array_free(array, TRUE);
 
-  infc_session_proxy_succeed_user_request(proxy, xml, user);
+  if(request != NULL)
+  {
+    infc_user_request_finished(INFC_USER_REQUEST(request), user);
+    infc_request_manager_remove_request(priv->request_manager, request);
+  }
 
   return TRUE;
 
@@ -968,7 +974,7 @@ infc_session_proxy_class_init(gpointer g_class,
   object_class->get_property = infc_session_proxy_get_property;
 
   proxy_class->translate_error = infc_session_proxy_translate_error_impl;
-  
+
   g_object_class_install_property(
     object_class,
     PROP_SESSION,
@@ -1069,8 +1075,10 @@ infc_session_proxy_get_type(void)
  * infc_session_proxy_set_connection:
  * @proxy: A #InfcSessionProxy.
  * @group: A #InfCommunicationJoinedGroup of subscribed connections. Ignored
- * if * @connection is %NULL.
+ * if @connection is %NULL.
  * @connection: A #InfXmlConnection.
+ * @seq_id: A sequence identifier for @connection. Ignored if @connection
+ * is %NULL.
  *
  * Sets the subscription connection for the given session. The subscription
  * connection is the connection through which session requests are transmitted
@@ -1081,6 +1089,12 @@ infc_session_proxy_get_type(void)
  * after synchronization. Note that no attempt is made to tell the other end
  * about the subscription.
  *
+ * @seq_id should be a sequence identifier obtained from the server side. It
+ * must be the same number that is used on the corresponding #InfdSessionProxy
+ * on the remote side, see infd_session_proxy_subscribe_to(). Normally
+ * #InfdDirectory or #InfcBrowser, respectively, take care of passing the
+ * correct sequence identifier.
+ *
  * When the subscription connection is being closed or replaced (by a
  * subsequent call to this function), all pending requests are dropped and
  * all users are set to be unavailable, but the session will not be closed,
@@ -1092,7 +1106,8 @@ infc_session_proxy_get_type(void)
 void
 infc_session_proxy_set_connection(InfcSessionProxy* proxy,
                                   InfCommunicationJoinedGroup* group,
-                                  InfXmlConnection* connection)
+                                  InfXmlConnection* connection,
+                                  guint seq_id)
 {
   InfcSessionProxyPrivate* priv;
   xmlNodePtr xml;
@@ -1145,6 +1160,9 @@ infc_session_proxy_set_connection(InfcSessionProxy* proxy,
     /* Set new group */
     priv->subscription_group = group;
     g_object_ref(priv->subscription_group);
+
+    g_assert(priv->request_manager == NULL);
+    priv->request_manager = infc_request_manager_new(seq_id);
   }
 
   inf_session_set_subscription_group(
@@ -1195,6 +1213,7 @@ infc_session_proxy_join_user(InfcSessionProxy* proxy,
   g_object_get(G_OBJECT(priv->session), "status", &status, NULL);
   g_return_val_if_fail(status == INF_SESSION_RUNNING, NULL);
   g_return_val_if_fail(priv->connection != NULL, NULL);
+  g_return_val_if_fail(priv->request_manager != NULL, NULL);
 
   /* TODO: Check params locally */
 
