@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /**
@@ -112,7 +113,7 @@ struct _InfDiscoveryAvahiPrivate {
   InfIo* io;
   InfXmppManager* xmpp_manager;
   InfXmppConnectionSecurityPolicy security_policy;
-  gnutls_certificate_credentials_t creds;
+  InfCertificateCredentials* creds;
   Gsasl* sasl_context;
   gchar* sasl_mechanisms;
 
@@ -287,6 +288,8 @@ inf_discovery_avahi_service_resolver_callback(AvahiServiceResolver* resolver,
 
   InfIpAddress* inf_addr;
   InfTcpConnection* tcp;
+  InfXmppConnection* xmpp;
+  InfXmlConnectionStatus status;
   GError* error;
   
   avahi = INF_DISCOVERY_AVAHI(userdata);
@@ -331,24 +334,20 @@ inf_discovery_avahi_service_resolver_callback(AvahiServiceResolver* resolver,
       break;
     }
 
-    discovery_info->resolved = inf_xmpp_manager_lookup_connection_by_address(
+    xmpp = inf_xmpp_manager_lookup_connection_by_address(
       priv->xmpp_manager,
       inf_addr,
       port
     );
 
-    if(discovery_info->resolved == NULL)
+    if(xmpp == NULL)
     {
-      error = NULL;
+      tcp = inf_tcp_connection_new(priv->io, inf_addr, port);
 
-      tcp = g_object_new(
-        INF_TYPE_TCP_CONNECTION,
-        "io", priv->io,
+      g_object_set(
+        G_OBJECT(tcp),
         "device-index", discovery_info->interface,
-        "remote-address", inf_addr,
-        "remote-port", port,
-        NULL
-      );
+        NULL);
 
       error = NULL;
       if(inf_tcp_connection_open(tcp, &error) == FALSE)
@@ -356,11 +355,11 @@ inf_discovery_avahi_service_resolver_callback(AvahiServiceResolver* resolver,
         inf_discovery_avahi_info_resolv_error(discovery_info, error);
         g_error_free(error);
 
-        g_object_unref(G_OBJECT(tcp));
+        g_object_unref(tcp);
       }
       else
       {
-        discovery_info->resolved = inf_xmpp_connection_new(
+        xmpp = inf_xmpp_connection_new(
           tcp,
           INF_XMPP_CONNECTION_CLIENT,
           NULL,
@@ -371,36 +370,60 @@ inf_discovery_avahi_service_resolver_callback(AvahiServiceResolver* resolver,
           priv->sasl_context == NULL ? NULL : priv->sasl_mechanisms
         );
 
-        g_object_unref(G_OBJECT(tcp));
+        g_object_unref(tcp);
 
-        inf_xmpp_manager_add_connection(
-          priv->xmpp_manager,
-          discovery_info->resolved
-        );
+        inf_xmpp_manager_add_connection(priv->xmpp_manager, xmpp);
 
-        inf_discovery_avahi_info_resolv_complete(discovery_info);
+        discovery_info->resolved = xmpp;
 
-        /* TODO: Also connect to notify::status and remove info->resolved when
-         * the connection is closed to guarantee to always return a connection
-         * that will soon be ready. */
         g_object_weak_ref(
-          G_OBJECT(discovery_info->resolved),
+          G_OBJECT(xmpp),
           inf_discovery_avahi_discovery_info_resolved_destroy_cb,
           discovery_info
         );
 
-        g_object_unref(G_OBJECT(discovery_info->resolved));
+        inf_discovery_avahi_info_resolv_complete(discovery_info);
+
+        g_object_unref(xmpp);
       }
     }
     else
     {
+      discovery_info->resolved = xmpp;
+
       g_object_weak_ref(
-        G_OBJECT(discovery_info->resolved),
+        G_OBJECT(xmpp),
         inf_discovery_avahi_discovery_info_resolved_destroy_cb,
         discovery_info
       );
 
-      inf_discovery_avahi_info_resolv_complete(discovery_info);
+      g_object_get(G_OBJECT(xmpp), "status", &status, NULL);
+
+      /* TODO: There is similar code in inf_discovery_avahi_resolve; should
+       * probably go into an extra function. */
+      if(status == INF_XML_CONNECTION_CLOSING)
+      {
+        /* TODO: That's a bit a sad case here. We should wait for the
+         * connection being closed, and then reopen it: */
+        inf_discovery_avahi_info_resolv_error(discovery_info, NULL);
+      }
+      else if(status == INF_XML_CONNECTION_CLOSED)
+      {
+        error = NULL;
+        if(!inf_xml_connection_open(INF_XML_CONNECTION(xmpp), &error))
+        {
+          inf_discovery_avahi_info_resolv_error(discovery_info, error);
+          g_error_free(error);
+        }
+        else
+        {
+          inf_discovery_avahi_info_resolv_complete(discovery_info);
+        }
+      }
+      else
+      {
+        inf_discovery_avahi_info_resolv_complete(discovery_info);
+      }
     }
     
     inf_ip_address_free(inf_addr);
@@ -770,7 +793,7 @@ inf_discovery_avahi_create_client(InfDiscoveryAvahi* discovery)
     AVAHI_CLIENT_NO_FAIL,
     inf_discovery_avahi_client_callback,
     discovery,
-    NULL
+    &error
   );
 
   /* This still seems to happen sometimes, even though we pass
@@ -1140,6 +1163,12 @@ inf_discovery_avahi_dispose(GObject* object)
     priv->xmpp_manager = NULL;
   }
 
+  if(priv->creds != NULL)
+  {
+    inf_certificate_credentials_unref(priv->creds);
+    priv->creds = NULL;
+  }
+
   if(priv->io != NULL)
   {
     g_object_unref(G_OBJECT(priv->io));
@@ -1186,12 +1215,10 @@ inf_discovery_avahi_set_property(GObject* object,
     priv->xmpp_manager = INF_XMPP_MANAGER(g_value_dup_object(value));
     break;
   case PROP_CREDENTIALS:
-    g_assert(priv->creds == NULL); /* construct only */
-    priv->creds =
-      (gnutls_certificate_credentials_t)g_value_get_pointer(value);
+    if(priv->creds != NULL) inf_certificate_credentials_unref(priv->creds);
+    priv->creds = (InfCertificateCredentials*)g_value_dup_boxed(value);
     break;
   case PROP_SASL_CONTEXT:
-    g_assert(priv->sasl_context == NULL); /* construct only */
     priv->sasl_context = (Gsasl*)g_value_get_pointer(value);
     break;
   case PROP_SASL_MECHANISMS:
@@ -1228,7 +1255,7 @@ inf_discovery_avahi_get_property(GObject* object,
     g_value_set_object(value, G_OBJECT(priv->xmpp_manager));
     break;
   case PROP_CREDENTIALS:
-    g_value_set_pointer(value, (gpointer)priv->creds);
+    g_value_set_boxed(value, priv->creds);
     break;
   case PROP_SASL_CONTEXT:
     g_value_set_pointer(value, (gpointer)priv->sasl_context);
@@ -1262,7 +1289,7 @@ inf_discovery_avahi_discover(InfDiscovery* discovery,
   if(info == NULL)
   {
     info = g_slice_new(InfDiscoveryAvahiDiscoverInfo);
-    
+
     info->type = g_strdup(type);
     info->service_browser = NULL;
     info->discovered = NULL;
@@ -1302,6 +1329,7 @@ inf_discovery_avahi_resolve(InfDiscovery* discovery,
 {
   InfDiscoveryAvahiPrivate* priv;
   InfDiscoveryAvahiInfoResolv* resolv;
+  InfXmlConnectionStatus status;
   int errno;
   GError* error;
 
@@ -1311,7 +1339,36 @@ inf_discovery_avahi_resolve(InfDiscovery* discovery,
 
   if(info->resolved != NULL)
   {
-    complete_func(info, INF_XML_CONNECTION(info->resolved), user_data);
+    g_object_get(G_OBJECT(info->resolved), "status", &status, NULL);
+
+    switch(status)
+    {
+    case INF_XML_CONNECTION_CLOSED:
+      error = NULL;
+      if(!inf_xml_connection_open(INF_XML_CONNECTION(info->resolved), &error))
+      {
+        error_func(info, error, user_data);
+        g_error_free(error);
+      }
+      else
+      {
+        complete_func(info, INF_XML_CONNECTION(info->resolved), user_data);
+      }
+
+      break;
+    case INF_XML_CONNECTION_CLOSING:
+      /* TODO: We should add ourselves to the resolver list, and wait for
+       * the connection being closed and reopen it afterwards. */
+      error_func(info, NULL, user_data);
+      break;
+    case INF_XML_CONNECTION_OPENING:
+    case INF_XML_CONNECTION_OPEN:
+      complete_func(info, INF_XML_CONNECTION(info->resolved), user_data);
+      break;
+    default:
+      g_assert_not_reached();
+      break;
+    }
   }
   else
   {
@@ -1335,7 +1392,7 @@ inf_discovery_avahi_resolve(InfDiscovery* discovery,
         inf_discovery_avahi_service_resolver_callback,
         discovery
       );
-      
+
       if(info->service_resolver == NULL)
       {
         error = NULL;
@@ -1393,7 +1450,7 @@ inf_discovery_avahi_publish(InfLocalPublisher* publisher,
 {
   InfDiscoveryAvahiPrivate* priv;
   InfLocalPublisherItem* item;
-  
+
   priv = INF_DISCOVERY_AVAHI_PRIVATE(publisher);
   item = g_slice_new(InfLocalPublisherItem);
 
@@ -1482,11 +1539,12 @@ inf_discovery_avahi_class_init(gpointer g_class,
   g_object_class_install_property(
     object_class,
     PROP_CREDENTIALS,
-    g_param_spec_pointer(
+    g_param_spec_boxed(
       "credentials",
       "Certificate credentials",
       "The GnuTLS certificate credentials used for encrypting XMPP streams",
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      INF_TYPE_CERTIFICATE_CREDENTIALS,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -1497,7 +1555,7 @@ inf_discovery_avahi_class_init(gpointer g_class,
       "sasl-context",
       "SASL context",
       "The Gsasl context used for authentication",
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -1509,7 +1567,7 @@ inf_discovery_avahi_class_init(gpointer g_class,
       "SASL mechanisms",
       "The accepted SASL mechanisms for authentication",
       NULL,
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -1631,6 +1689,10 @@ inf_discovery_avahi_get_type(void)
  * These may be %NULL in which case #InfXmppConnection uses builtin
  * credentials or a builtin context, respectively.
  *
+ * If this #InfDiscoveryAvahi is not used to discover services but only to
+ * publish any, then @creds, @sasl_context and @sasl_mechanisms are ignored
+ * and can safely set to be %NULL.
+ *
  * @sasl_mechanisms specifies allowed mechanisms used for authentication with
  * the server. It can be %NULL, in which case all available mechanisms are
  * accepted.
@@ -1640,7 +1702,7 @@ inf_discovery_avahi_get_type(void)
 InfDiscoveryAvahi*
 inf_discovery_avahi_new(InfIo* io,
                         InfXmppManager* manager,
-                        gnutls_certificate_credentials_t creds,
+                        InfCertificateCredentials* creds,
                         Gsasl* sasl_context,
                         const gchar* sasl_mechanisms)
 {
