@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfinity/common/inf-standalone-io.h>
@@ -443,6 +444,11 @@ inf_standalone_io_io_watch(InfIo* io,
       priv->watches,
       priv->fd_alloc * sizeof(InfStandaloneIoWatch)
     );
+
+    /* Update event pointers, the location of the events in memory might have
+     * changed after realloc. */
+    for(i = 0; i < priv->fd_size; ++i)
+      priv->watches[i].event = &priv->events[i];
   }
 
 #ifdef G_OS_WIN32
@@ -588,6 +594,9 @@ inf_standalone_io_get_type(void)
  * inf_standalone_io_new:
  *
  * Creates a new #InfStandaloneIo.
+ *
+ * Returns: A new #InfStandaloneIo. Free with g_object_unref() when no longer
+ * needed.
  **/
 InfStandaloneIo*
 inf_standalone_io_new(void)
