@@ -13,10 +13,12 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfinity/common/inf-tcp-connection.h>
+#include <libinfinity/common/inf-tcp-connection-private.h>
 #include <libinfinity/common/inf-ip-address.h>
 #include <libinfinity/common/inf-io.h>
 #include <libinfinity/inf-marshal.h>
@@ -242,7 +244,7 @@ inf_tcp_connection_send_real(InfTcpConnection* connection,
     }
     else if(result > 0)
     {
-      send_data = (char*)send_data + result;
+      send_data = (const char*)send_data + result;
       send_len -= result;
     }
   } while( (send_len > 0) &&
@@ -652,7 +654,7 @@ inf_tcp_connection_get_property(GObject* object,
           priv->device_index,
           strerror(errno)
         );
-        
+
         g_value_set_string(value, NULL);
       }
       else
@@ -675,7 +677,7 @@ inf_tcp_connection_error(InfTcpConnection* connection,
   InfTcpConnectionPrivate* priv;
   priv = INF_TCP_CONNECTION_PRIVATE(connection);
 
-  /* Normally, it would be enough to check one both conditions, but socket
+  /* Normally, it would be enough to check one of both conditions, but socket
    * may be already set with status still being CLOSED during
    * inf_tcp_connection_open(). */
   if(priv->events != 0)
@@ -951,6 +953,77 @@ inf_tcp_connection_get_type(void)
   return tcp_connection_type;
 }
 
+/**
+ * inf_tcp_connection_new:
+ * @io: A #InfIo object used to watch for activity.
+ * @remote_addr: The address to eventually connect to.
+ * @remote_port: The port to eventually connect to.
+ *
+ * Creates a new #InfTcpConnection. The arguments are stored as properties for
+ * an eventual inf_tcp_connection_open() call, this function itself does not
+ * establish a connection.
+ *
+ * Returns: A new #InfTcpConnection. Free with g_object_unref().
+ **/
+InfTcpConnection*
+inf_tcp_connection_new(InfIo* io,
+                       InfIpAddress* remote_addr,
+                       guint remote_port)
+{
+  InfTcpConnection* tcp;
+
+  g_return_val_if_fail(INF_IS_IO(io), NULL);
+  g_return_val_if_fail(remote_addr != NULL, NULL);
+  g_return_val_if_fail(remote_port <= 65535, NULL);
+
+  tcp = INF_TCP_CONNECTION(
+    g_object_new(
+      INF_TYPE_TCP_CONNECTION,
+      "io", io,
+      "remote-address", remote_addr,
+      "remote-port", remote_port,
+      NULL));
+
+  return tcp;
+}
+
+/**
+ * inf_tcp_connection_new_and_open:
+ * @io: A #InfIo object used to watch for activity.
+ * @remote_addr: The address to connect to.
+ * @remote_port: The port to connect to.
+ * @error: Location to store error information.
+ *
+ * Creates a new #InfTcpConnection and connects it to the given TCP endpoint.
+ * Like inf_tcp_connection_new, but calls inf_tcp_connection_open().
+ *
+ * Returns: A new #InfTcpConnection, or %NULL on error. Free with
+ * g_object_unref().
+ **/
+InfTcpConnection*
+inf_tcp_connection_new_and_open(InfIo* io,
+                                InfIpAddress* remote_addr,
+                                guint remote_port,
+                                GError** error)
+{
+  InfTcpConnection* tcp;
+
+  g_return_val_if_fail(INF_IS_IO(io), NULL);
+  g_return_val_if_fail(remote_addr != NULL, NULL);
+  g_return_val_if_fail(remote_port <= 65535, NULL);
+  g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+  tcp = inf_tcp_connection_new(io, remote_addr, remote_port);
+
+  if(inf_tcp_connection_open(tcp, error) == FALSE)
+  {
+    g_object_unref(tcp);
+    return NULL;
+  }
+
+  return tcp;
+}
+
 /**
  * inf_tcp_connection_open:
  * @connection: A #InfTcpConnection.
@@ -958,7 +1031,7 @@ inf_tcp_connection_get_type(void)
  *
  * Attempts to open @connection. Make sure to have set the "remote-address"
  * and "remote-port" property before calling this function. If an error
- * occurs, the function returns FALSE and @error is set. Note however that
+ * occurs, the function returns %FALSE and @error is set. Note however that
  * the connection might not be fully open when the function returns
  * (check the "status" property if you need to know). If an asynchronous
  * error occurs while the connection is being opened, the "error" signal
@@ -996,7 +1069,7 @@ inf_tcp_connection_open(InfTcpConnection* connection,
   g_return_val_if_fail(priv->remote_port != 0, FALSE);
 
   /* Close previous socket */
-  if(priv->socket == INVALID_SOCKET)
+  if(priv->socket != INVALID_SOCKET)
     closesocket(priv->socket);
 
   switch(inf_ip_address_get_family(priv->remote_address))
@@ -1203,7 +1276,7 @@ inf_tcp_connection_send(InfTcpConnection* connection,
 
     if(inf_tcp_connection_send_real(connection, data, &sent_len) == TRUE)
     {
-      data = (char*)data + sent_len;
+      data = (const char*)data + sent_len;
       len -= sent_len;
     }
     else
@@ -1368,15 +1441,7 @@ _inf_tcp_connection_accepted(InfIo* io,
   g_return_val_if_fail(address != NULL, NULL);
   g_return_val_if_fail(port != 0, NULL);
 
-  connection = INF_TCP_CONNECTION(
-    g_object_new(
-      INF_TYPE_TCP_CONNECTION,
-      "io", io,
-      "remote-address", address,
-      "remote-port", port,
-      NULL
-    )
-  );
+  connection = inf_tcp_connection_new(io, address, port);
 
   inf_ip_address_free(address);
 
