@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /**
@@ -31,6 +32,7 @@
 #include <libinfinity/communication/inf-communication-central-method.h>
 #include <libinfinity/communication/inf-communication-hosted-group.h>
 #include <libinfinity/communication/inf-communication-registry.h>
+#include <libinfinity/inf-signals.h>
 
 #include <libxml/xmlsave.h>
 
@@ -64,7 +66,6 @@ inf_communication_central_method_notify_status_cb(GObject* object,
   InfCommunicationCentralMethod* method;
   InfCommunicationCentralMethodPrivate* priv;
   InfXmlConnectionStatus status;
-  gpointer was_open_ptr;
 
   method = INF_COMMUNICATION_CENTRAL_METHOD(user_data);
   priv = INF_COMMUNICATION_CENTRAL_METHOD_PRIVATE(method);
@@ -164,7 +165,7 @@ inf_communication_central_method_remove_member(InfCommunicationMethod* method,
     );
   }
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     connection,
     G_CALLBACK(inf_communication_central_method_notify_status_cb),
     method
@@ -203,46 +204,27 @@ inf_communication_central_method_send_single(InfCommunicationMethod* method,
 
 static void
 inf_communication_central_method_send_all(InfCommunicationMethod* method,
-                                          InfXmlConnection* except,
                                           xmlNodePtr xml)
 {
   InfCommunicationCentralMethodPrivate* priv;
   GSList* item;
   GSList* next;
-  gboolean drop_xml;
 
   priv = INF_COMMUNICATION_CENTRAL_METHOD_PRIVATE(method);
 
-  drop_xml = FALSE;
   for(item = priv->connections; item != NULL; item = next)
   {
     next = item->next;
-    if(item->data == except) continue;
 
-    if(next == NULL || (next->data == except && next->next == NULL))
-    {
-      g_assert(drop_xml == FALSE);
-      drop_xml = TRUE;
-
-      inf_communication_registry_send(
-        priv->registry,
-        priv->group,
-        INF_XML_CONNECTION(item->data),
-        xml
-      );
-    }
-    else
-    {
-      inf_communication_registry_send(
-        priv->registry,
-        priv->group,
-        INF_XML_CONNECTION(item->data),
-        xmlCopyNode(xml, 1)
-      );
-    }
+    inf_communication_registry_send(
+      priv->registry,
+      priv->group,
+      INF_XML_CONNECTION(item->data),
+      next ? xmlCopyNode(xml, 1) : xml
+    );
   }
 
-  if(!drop_xml)
+  if(priv->connections == NULL)
     xmlFreeNode(xml);
 }
 
@@ -273,6 +255,7 @@ inf_communication_central_method_received(InfCommunicationMethod* method,
   xmlSaveCtxtPtr ctx;
   gchar* remote_id;
   gchar* publisher_id;
+  GSList* item;
 
   priv = INF_COMMUNICATION_CENTRAL_METHOD_PRIVATE(method);
   error = NULL;
@@ -319,11 +302,18 @@ inf_communication_central_method_received(InfCommunicationMethod* method,
 
     if(priv->is_publisher && scope == INF_COMMUNICATION_SCOPE_GROUP)
     {
-      inf_communication_central_method_send_all(
-        method,
-        connection,
-        xmlCopyNode(xml, 1)
-      );
+      for(item = priv->connections; item != NULL; item = item->next)
+      {
+        if(item->data != connection)
+        {
+          inf_communication_registry_send(
+            priv->registry,
+            priv->group,
+            INF_XML_CONNECTION(item->data),
+            xmlCopyNode(xml, 1)
+          );
+        }
+      }
 
       /* TODO: Forward to other methods of same group */
     }
