@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /**
@@ -37,6 +38,7 @@
 
 #include <libinfinity/communication/inf-communication-registry.h>
 #include <libinfinity/common/inf-xml-util.h>
+#include <libinfinity/inf-signals.h>
 
 #include <string.h>
 
@@ -573,19 +575,19 @@ inf_communication_registry_remove_connection(InfCommunicationRegistry* rgstry,
   }
   else
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(connection),
       G_CALLBACK(inf_communication_registry_received_cb),
       rgstry
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(connection),
       G_CALLBACK(inf_communication_registry_sent_cb),
       rgstry
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(connection),
       G_CALLBACK(inf_communication_registry_notify_status_cb),
       rgstry
@@ -694,19 +696,19 @@ inf_communication_registry_dispose(GObject* object)
     g_hash_table_iter_init(&iter, priv->connections);
     while(g_hash_table_iter_next(&iter, &key, NULL))
     {
-      g_signal_handlers_disconnect_by_func(
+      inf_signal_handlers_disconnect_by_func(
         G_OBJECT(key),
         G_CALLBACK(inf_communication_registry_received_cb),
         registry
       );
 
-      g_signal_handlers_disconnect_by_func(
+      inf_signal_handlers_disconnect_by_func(
         G_OBJECT(key),
         G_CALLBACK(inf_communication_registry_sent_cb),
         registry
       );
 
-      g_signal_handlers_disconnect_by_func(
+      inf_signal_handlers_disconnect_by_func(
         G_OBJECT(key),
         G_CALLBACK(inf_communication_registry_notify_status_cb),
         registry
