@@ -1,3 +1,4 @@
+#include <libinfinity/inf-marshal.h>
 
 #include	<glib-object.h>
 
@@ -81,7 +82,7 @@ inf_marshal_VOID__BOXED_DOUBLE_DOUBLE (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__BOXED_DOUBLE_DOUBLE) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__BOXED_DOUBLE_DOUBLE*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_boxed (param_values + 1),
@@ -122,7 +123,7 @@ inf_marshal_VOID__UINT_UINT_OBJECT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__UINT_UINT_OBJECT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__UINT_UINT_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_uint (param_values + 1),
@@ -162,7 +163,7 @@ inf_marshal_VOID__OBJECT_DOUBLE (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__OBJECT_DOUBLE) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__OBJECT_DOUBLE*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_object (param_values + 1),
@@ -199,7 +200,7 @@ inf_marshal_VOID__OBJECT_POINTER (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__OBJECT_POINTER) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__OBJECT_POINTER*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_object (param_values + 1),
@@ -236,7 +237,7 @@ inf_marshal_VOID__POINTER_UINT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__POINTER_UINT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__POINTER_UINT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_pointer (param_values + 1),
@@ -275,7 +276,7 @@ inf_marshal_VOID__UINT_UINT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__UINT_UINT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__UINT_UINT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_uint (param_values + 1),
@@ -312,7 +313,7 @@ inf_marshal_VOID__OBJECT_BOOLEAN (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__OBJECT_BOOLEAN) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__OBJECT_BOOLEAN*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_object (param_values + 1),
@@ -350,7 +351,7 @@ inf_marshal_VOID__UINT_BOXED_OBJECT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__UINT_BOXED_OBJECT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__UINT_BOXED_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_uint (param_values + 1),
@@ -388,7 +389,7 @@ inf_marshal_VOID__UINT_INT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__UINT_INT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__UINT_INT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_uint (param_values + 1),
@@ -425,7 +426,7 @@ inf_marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__OBJECT_OBJECT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__OBJECT_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_object (param_values + 1),
@@ -462,7 +463,7 @@ inf_marshal_VOID__BOXED_OBJECT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__BOXED_OBJECT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__BOXED_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_boxed (param_values + 1),
@@ -500,7 +501,7 @@ inf_marshal_VOID__BOXED_BOXED_OBJECT (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__BOXED_BOXED_OBJECT) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__BOXED_BOXED_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_boxed (param_values + 1),
@@ -539,7 +540,7 @@ inf_marshal_VOID__OBJECT_OBJECT_BOOLEAN (GClosure     *closure,
       data1 = g_value_peek_pointer (param_values + 0);
       data2 = closure->data;
     }
-  callback = (GMarshalFunc_VOID__OBJECT_OBJECT_BOOLEAN) (marshal_data ? marshal_data : cc->callback);
+  callback = *(GMarshalFunc_VOID__OBJECT_OBJECT_BOOLEAN*) (marshal_data ? &marshal_data : &cc->callback);
 
   callback (data1,
             g_marshal_value_peek_object (param_values + 1),
@@ -552,3 +553,81 @@ inf_marshal_VOID__OBJECT_OBJECT_BOOLEAN (GClosure     *closure,
 
 /* VOID:ENUM (inf-marshal.in:20) */
 
+/* BOOLEAN:OBJECT,OBJECT (inf-marshal.in:21) */
+void
+inf_marshal_BOOLEAN__OBJECT_OBJECT (GClosure     *closure,
+                                    GValue       *return_value G_GNUC_UNUSED,
+                                    guint         n_param_values,
+                                    const GValue *param_values,
+                                    gpointer      invocation_hint G_GNUC_UNUSED,
+                                    gpointer      marshal_data)
+{
+  typedef gboolean (*GMarshalFunc_BOOLEAN__OBJECT_OBJECT) (gpointer     data1,
+                                                           gpointer     arg_1,
+                                                           gpointer     arg_2,
+                                                           gpointer     data2);
+  register GMarshalFunc_BOOLEAN__OBJECT_OBJECT callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+  gboolean v_return;
+
+  g_return_if_fail (return_value != NULL);
+  g_return_if_fail (n_param_values == 3);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = *(GMarshalFunc_BOOLEAN__OBJECT_OBJECT*) (marshal_data ? &marshal_data : &cc->callback);
+
+  v_return = callback (data1,
+                       g_marshal_value_peek_object (param_values + 1),
+                       g_marshal_value_peek_object (param_values + 2),
+                       data2);
+
+  g_value_set_boolean (return_value, v_return);
+}
+
+/* VOID:OBJECT,UINT (inf-marshal.in:22) */
+void
+inf_marshal_VOID__OBJECT_UINT (GClosure     *closure,
+                               GValue       *return_value G_GNUC_UNUSED,
+                               guint         n_param_values,
+                               const GValue *param_values,
+                               gpointer      invocation_hint G_GNUC_UNUSED,
+                               gpointer      marshal_data)
+{
+  typedef void (*GMarshalFunc_VOID__OBJECT_UINT) (gpointer     data1,
+                                                  gpointer     arg_1,
+                                                  guint        arg_2,
+                                                  gpointer     data2);
+  register GMarshalFunc_VOID__OBJECT_UINT callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+
+  g_return_if_fail (n_param_values == 3);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = *(GMarshalFunc_VOID__OBJECT_UINT*) (marshal_data ? &marshal_data : &cc->callback);
+
+  callback (data1,
+            g_marshal_value_peek_object (param_values + 1),
+            g_marshal_value_peek_uint (param_values + 2),
+            data2);
+}
+
