@@ -13,9 +13,15 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
+#ifndef G_OS_WIN32
+/* Need this for O_NOFOLLOW and fdopendir and so on */
+# define _GNU_SOURCE
+#endif
+
 #include <libinfinity/server/infd-filesystem-storage.h>
 #include <libinfinity/server/infd-storage.h>
 #include <libinfinity/inf-i18n.h>
@@ -25,6 +31,16 @@
 #include <string.h>
 #include <errno.h>
 
+#ifdef G_OS_WIN32
+# include <windows.h>
+#else
+# include <sys/types.h>
+# include <sys/stat.h>
+# include <fcntl.h>
+# include <dirent.h>
+# include <unistd.h>
+#endif
+
 typedef struct _InfdFilesystemStoragePrivate InfdFilesystemStoragePrivate;
 struct _InfdFilesystemStoragePrivate {
   gchar* root_directory;
@@ -53,12 +69,14 @@ infd_filesystem_storage_verify_path(const gchar* path,
   components = g_strsplit(path, "/", 0);
   for(component = components; *component != NULL; ++ component)
   {
-    if(strcmp(*component, ".") == 0 || strcmp(*component, "..") == 0)
+    if(*component == '\0' ||
+       strcmp(*component, ".") == 0 || strcmp(*component, "..") == 0)
     {
       g_set_error(
         error,
         infd_filesystem_storage_error_quark,
         INFD_FILESYSTEM_STORAGE_ERROR_INVALID_PATH,
+        "%s",
         _("The path contains invalid components")
       );
 
@@ -118,6 +136,8 @@ infd_filesystem_storage_system_error(int code,
 {
   /* TODO_Win32: Use FormatMessage or something on Win32,
    * or probably better g_win32_error_message(). */
+  /* TODO: Actually we should get away from including system error codes
+   * in these errors, as they get eventually sent over the net */
   g_set_error(
     error,
     infd_filesystem_storage_system_error_quark,
@@ -131,6 +151,47 @@ static gboolean
 infd_filesystem_storage_remove_rec(const gchar* path,
                                    GError** error)
 {
+#ifdef G_OS_WIN32
+  SHFILEOPSTRUCTW op;
+  gchar* from_dup;
+  gunichar2* from;
+  glong len;
+  gboolean result;
+  int error_code;
+
+  from = g_utf8_to_utf16(path, -1, NULL, &len, error);
+  if(!from) return FALSE;
+
+  from = g_realloc(from, (len+2)*sizeof(gunichar2));
+  from[len+1] = L'\0';
+
+  op.hwnd = NULL;
+  op.wFunc = FO_DELETE;
+  op.pFrom = from;
+  op.pTo = NULL;
+  op.fFlags = FOF_SILENT | FOF_NOERRORUI | FOF_NOCONFIRMATION;
+  op.fAnyOperationsAborted = 0;
+  op.hNameMappings = NULL;
+  op.lpszProgressTitle = NULL;
+
+  result = TRUE;
+  error_code = SHFileOperationW(&op);
+  if(error_code != 0 || op.fAnyOperationsAborted != 0)
+  {
+    g_set_error(
+      error,
+      infd_filesystem_storage_error_quark,
+      INFD_FILESYSTEM_STORAGE_ERROR_REMOVE_FILES,
+      "Failed to remove files from disk"
+    );
+
+    result = FALSE;
+  }
+
+  g_free(from);
+  return result;
+#else
+  /* TODO: Use the REMOVE_FILES error code when something fails below */
   GDir* dir;
   const gchar* name;
   gchar* child;
@@ -175,6 +236,7 @@ infd_filesystem_storage_remove_rec(const gchar* path,
   }
 
   return TRUE;
+#endif
 }
 
 static void
@@ -262,8 +324,17 @@ infd_filesystem_storage_storage_read_subdirectory(InfdStorage* storage,
   InfdFilesystemStorage* fs_storage;
   InfdFilesystemStoragePrivate* priv;
   GSList* list;
+#ifndef G_OS_WIN32
+  int dir_fd;
+  DIR* dir;
+  struct dirent* dir_entry;
+  struct dirent* dir_result;
+  struct stat stat_buf;
+  int saved_errno;
+#else
   GDir* dir;
   const gchar* name;
+#endif
   gchar* converted_name;
   gchar* full_name;
   gchar* file_path;
@@ -283,12 +354,91 @@ infd_filesystem_storage_storage_read_subdirectory(InfdStorage* storage,
   full_name = g_build_filename(priv->root_directory, converted_name, NULL);
   g_free(converted_name);
 
+  list = NULL;
+
+#ifndef G_OS_WIN32
+  dir_fd = open(full_name, O_NOFOLLOW | O_RDONLY);
+  if(dir_fd == -1 || (dir = fdopendir(dir_fd)) == NULL)
+  {
+    infd_filesystem_storage_system_error(errno, error);
+    g_free(full_name);
+    return NULL;
+  }
+
+  dir_entry = g_malloc(offsetof(struct dirent, d_name) +
+                       fpathconf(dir_fd, _PC_NAME_MAX) + 1);
+
+  for(saved_errno = readdir_r(dir, dir_entry, &dir_result);
+      saved_errno == 0 && dir_result != NULL;
+      saved_errno = readdir_r(dir, dir_entry, &dir_result))
+  {
+    converted_name = g_filename_to_utf8(dir_result->d_name,
+                                        -1,
+                                        NULL,
+                                        &name_len,
+                                        NULL);
+    if(converted_name != NULL && strcmp(converted_name, ".") != 0
+                              && strcmp(converted_name, "..") != 0)
+    {
+      /* Some filesystems, such as reiserfs, don't support reporting the
+       * entry's file type. In that case we do an additional lstat here. */
+      if(dir_result->d_type == DT_UNKNOWN)
+      {
+        file_path = g_build_filename(full_name, dir_result->d_name, NULL);
+        if(lstat(file_path, &stat_buf) == 0)
+        {
+          if(S_ISDIR(stat_buf.st_mode))
+            dir_result->d_type = DT_DIR;
+          else if(S_ISREG(stat_buf.st_mode))
+            dir_result->d_type = DT_REG;
+          else if(S_ISLNK(stat_buf.st_mode))
+            dir_result->d_type = DT_LNK;
+        }
+        g_free(file_path);
+      }
+
+      if(dir_result->d_type != DT_LNK && dir_result->d_type != DT_UNKNOWN)
+      {
+        if(dir_result->d_type == DT_DIR)
+        {
+          list = g_slist_prepend(
+            list,
+            infd_storage_node_new_subdirectory(converted_name)
+          );
+        }
+        else if(dir_result->d_type == DT_REG)
+        {
+          /* The note type identifier is behind the last '.' */
+          separator = g_strrstr_len(converted_name, name_len, ".");
+          if(separator != NULL)
+          {
+            *separator = '\0';
+            list = g_slist_prepend(
+              list,
+              infd_storage_node_new_note(converted_name, separator + 1)
+            );
+          }
+        }
+      }
+    }
+
+    g_free(converted_name);
+  }
+
+  g_free(dir_entry);
+
+  if(saved_errno != 0)
+  {
+    infd_filesystem_storage_system_error(saved_errno, error);
+    g_free(full_name);
+    return NULL;
+  }
+#else
   dir = g_dir_open(full_name, 0, error);
 
   if(dir == NULL)
     return NULL;
 
-  list = NULL;
   for(name = g_dir_read_name(dir); name != NULL; name = g_dir_read_name(dir))
   {
     converted_name = g_filename_to_utf8(name, -1, NULL, &name_len, NULL);
@@ -322,6 +472,7 @@ infd_filesystem_storage_storage_read_subdirectory(InfdStorage* storage,
   }
 
   g_dir_close(dir);
+#endif
   g_free(full_name);
   return list;
 }
@@ -376,6 +527,9 @@ infd_filesystem_storage_storage_remove_node(InfdStorage* storage,
   gchar* converted_name;
   gchar* disk_name;
   gchar* full_name;
+#ifdef G_OS_WIN32
+  gchar* sep;
+#endif
   gboolean ret;
 
   fs_storage = INFD_FILESYSTEM_STORAGE(storage);
@@ -388,6 +542,14 @@ infd_filesystem_storage_storage_remove_node(InfdStorage* storage,
   if(converted_name == NULL)
     return FALSE;
 
+#ifdef G_OS_WIN32
+  /* This is required for SHFileOperation. Maybe we should do it at a central
+   * place also for the other file operations... */
+  for(sep = converted_name; *sep != '\0'; ++sep)
+    if(*sep == '/')
+      *sep = '\\';
+#endif
+
   if(identifier != NULL)
   {
     disk_name = g_strconcat(converted_name, ".", identifier, NULL);
@@ -529,7 +691,8 @@ infd_filesystem_storage_new(const gchar* root_directory)
 
 /**
  * infd_filesystem_storage_open:
- * @storage:  A #InfdFilesystemStorage.
+ * @storage: A #InfdFilesystemStorage.
+ * @identifier: The type of node to open.
  * @path: Tha path to open.
  * @mode: Either "r" for reading or "w" for writing.
  * @error: Location to store error information, if any.
@@ -552,14 +715,18 @@ infd_filesystem_storage_open(InfdFilesystemStorage* storage,
   gchar* full_name;
   FILE* res;
   int save_errno;
+#ifndef G_OS_WIN32
+  int fd;
+  int open_mode;
+#endif
 
   priv = INFD_FILESYSTEM_STORAGE_PRIVATE(storage);
   if(infd_filesystem_storage_verify_path(path, error) == FALSE)
-    return FALSE;
+    return NULL;
 
   converted_name = g_filename_from_utf8(path, -1, NULL, NULL, error);
   if(converted_name == NULL)
-    return FALSE;
+    return NULL;
 
   disk_name = g_strconcat(converted_name, ".", identifier, NULL);
   g_free(converted_name);
@@ -567,14 +734,25 @@ infd_filesystem_storage_open(InfdFilesystemStorage* storage,
   full_name = g_build_filename(priv->root_directory, disk_name, NULL);
   g_free(disk_name);
 
+#ifdef G_OS_WIN32
   res = g_fopen(full_name, mode);
+#else
+  if(strcmp(mode, "r") == 0) open_mode = O_RDONLY;
+  else if(strcmp(mode, "w") == 0) open_mode = O_CREAT | O_WRONLY | O_TRUNC;
+  else g_assert_not_reached();
+  fd = open(full_name, O_NOFOLLOW | open_mode, 0600);
+  if(fd == -1)
+    res = NULL;
+  else
+    res = fdopen(fd, mode);
+#endif
   save_errno = errno;
   g_free(full_name);
 
   if(res == NULL)
   {
     infd_filesystem_storage_system_error(save_errno, error);
-    return FALSE;
+    return NULL;
   }
 
   return res;
