@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinfinity/server/infd-xmpp-server.h>
@@ -21,6 +22,14 @@
 #include <libinfinity/server/infd-xml-server.h>
 #include <libinfinity/common/inf-xmpp-connection.h>
 #include <libinfinity/inf-marshal.h>
+#include <libinfinity/inf-signals.h>
+
+/* Some Windows header #defines ERROR for no good */
+#ifdef G_OS_WIN32
+# ifdef ERROR
+#  undef ERROR
+# endif
+#endif
 
 typedef enum InfdXmppServerStatus_ {
   INFD_XMPP_SERVER_CLOSED,
@@ -34,7 +43,7 @@ struct _InfdXmppServerPrivate {
   gchar* local_hostname;
   InfXmppConnectionSecurityPolicy security_policy;
 
-  gnutls_certificate_credentials_t tls_creds;
+  InfCertificateCredentials* tls_creds;
 
   Gsasl* sasl_context;
   Gsasl* sasl_own_context;
@@ -141,6 +150,7 @@ infd_xmpp_server_notify_status_cb(InfdTcpServer* tcp_server,
   switch(status)
   {
   case INFD_TCP_SERVER_CLOSED:
+  case INFD_TCP_SERVER_BOUND:
     if(priv->status != INFD_XMPP_SERVER_CLOSED)
     {
       priv->status = INFD_XMPP_SERVER_CLOSED;
@@ -179,20 +189,20 @@ infd_xmpp_server_set_tcp(InfdXmppServer* xmpp,
     /* This will cause a notify that will adjust the XMPP status later */
     if(tcp_status != INFD_TCP_SERVER_CLOSED)
       infd_tcp_server_close(priv->tcp);
-    
-    g_signal_handlers_disconnect_by_func(
+
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->tcp),
       G_CALLBACK(infd_xmpp_server_new_connection_cb),
       xmpp
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->tcp),
       G_CALLBACK(infd_xmpp_server_error_cb),
       xmpp
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(priv->tcp),
       G_CALLBACK(infd_xmpp_server_notify_status_cb),
       xmpp
@@ -233,6 +243,7 @@ infd_xmpp_server_set_tcp(InfdXmppServer* xmpp,
     switch(tcp_status)
     {
     case INFD_TCP_SERVER_CLOSED:
+    case INFD_TCP_SERVER_BOUND:
       g_assert(priv->status == INFD_XMPP_SERVER_CLOSED);
       break;
     case INFD_TCP_SERVER_OPEN:
@@ -273,6 +284,35 @@ infd_xmpp_server_sasl_cb(Gsasl* ctx,
   }
 }
 
+/* Set own SASL context based on whether an external one is given or not */
+static void
+infd_xmpp_server_setup_own_sasl_context(InfdXmppServer* xmpp)
+{
+  InfdXmppServerPrivate* priv;
+  priv = INFD_XMPP_SERVER_PRIVATE(xmpp);
+
+  if(priv->sasl_context == NULL && priv->sasl_own_context == NULL)
+  {
+    /* Failure does not matter too much because every XMPP connection will
+     * generate an own SASL context in this case. */
+    if(gsasl_init(&priv->sasl_own_context) == GSASL_OK)
+    {
+      priv->sasl_context = priv->sasl_own_context;
+      gsasl_callback_set(priv->sasl_context, infd_xmpp_server_sasl_cb);
+      gsasl_callback_hook_set(priv->sasl_context, xmpp);
+      g_object_notify(G_OBJECT(xmpp), "sasl-context");
+    }
+  }
+  else if(priv->sasl_context != NULL && priv->sasl_own_context != NULL &&
+          priv->sasl_context != priv->sasl_own_context)
+  {
+    /* TODO: Make sure that the Gsasl context is no longer in use by a XMPP
+     * connection object. */
+    gsasl_done(priv->sasl_own_context);
+    priv->sasl_own_context = NULL;
+  }
+}
+
 static void
 infd_xmpp_server_init(GTypeInstance* instance,
                       gpointer g_class)
@@ -286,7 +326,7 @@ infd_xmpp_server_init(GTypeInstance* instance,
   priv->tcp = NULL;
   priv->status = INFD_XMPP_SERVER_CLOSED;
   priv->local_hostname = g_strdup(g_get_host_name());
-  priv->security_policy = INF_XMPP_CONNECTION_SECURITY_BOTH_PREFER_TLS;
+  priv->security_policy = INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED;
 
   priv->tls_creds = NULL;
   priv->sasl_context = NULL;
@@ -316,19 +356,7 @@ infd_xmpp_server_constructor(GType type,
     priv->tls_creds != NULL
   );
 
-  if(priv->sasl_context == NULL)
-  {
-    /* Failure does not matter too much because every XMPP connection will
-     * generate an own SASL context in this case. */
-    if(gsasl_init(&priv->sasl_own_context) == GSASL_OK)
-    {
-      priv->sasl_context = priv->sasl_own_context;
-      gsasl_callback_set(priv->sasl_context, infd_xmpp_server_sasl_cb);
-      gsasl_callback_hook_set(priv->sasl_context, obj);
-      g_object_notify(G_OBJECT(obj), "sasl-context");
-    }
-  }
-
+  infd_xmpp_server_setup_own_sasl_context(INFD_XMPP_SERVER(obj));
   return obj;
 }
 
@@ -345,7 +373,16 @@ infd_xmpp_server_dispose(GObject* object)
     infd_xml_server_close(INFD_XML_SERVER(xmpp));
 
   if(priv->tcp != NULL)
+  {
     g_object_unref(G_OBJECT(priv->tcp));
+    priv->tcp = NULL;
+  }
+
+  if(priv->tls_creds != NULL)
+  {
+    inf_certificate_credentials_unref(priv->tls_creds);
+    priv->tls_creds = NULL;
+  }
 
   G_OBJECT_CLASS(parent_class)->dispose(object);
 }
@@ -397,25 +434,20 @@ infd_xmpp_server_set_property(GObject* object,
       priv->local_hostname = g_strdup(g_get_host_name());
     break;
   case PROP_CREDENTIALS:
-    priv->tls_creds = g_value_get_pointer(value);
+    if(priv->tls_creds != NULL)
+      inf_certificate_credentials_unref(priv->tls_creds);
+    priv->tls_creds = g_value_dup_boxed(value);
     break;
   case PROP_SASL_CONTEXT:
-    /* TODO: Make sure that the Gsasl context is no longer in use by a XMPP
-     * connection object. */
-    if(priv->sasl_own_context != NULL)
-    {
-      gsasl_done(priv->sasl_own_context);
-      priv->sasl_own_context = NULL;
-    }
-    
     priv->sasl_context = g_value_get_pointer(value);
+    infd_xmpp_server_setup_own_sasl_context(xmpp);
     break;
   case PROP_SASL_MECHANISMS:
     g_free(priv->sasl_mechanisms);
     priv->sasl_mechanisms = g_value_dup_string(value);
     break;
   case PROP_SECURITY_POLICY:
-    priv->security_policy = g_value_get_enum(value);
+    infd_xmpp_server_set_security_policy(xmpp, g_value_get_enum(value));
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -459,7 +491,7 @@ infd_xmpp_server_get_property(GObject* object,
     g_value_set_string(value, priv->local_hostname);
     break;
   case PROP_CREDENTIALS:
-    g_value_set_pointer(value, priv->tls_creds);
+    g_value_set_boxed(value, priv->tls_creds);
     break;
   case PROP_SASL_CONTEXT:
     g_value_set_pointer(value, priv->sasl_context);
@@ -545,11 +577,12 @@ infd_xmpp_server_class_init(gpointer g_class,
   g_object_class_install_property(
     object_class,
     PROP_CREDENTIALS,
-    g_param_spec_pointer(
+    g_param_spec_boxed(
       "credentials",
       "Credentials",
       "The certificate credentials for GnuTLS",
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      INF_TYPE_CERTIFICATE_CREDENTIALS,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -560,7 +593,7 @@ infd_xmpp_server_class_init(gpointer g_class,
       "sasl-context",
       "GnuSASL context",
       "The GnuSASL context used for authentaction",
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -572,7 +605,7 @@ infd_xmpp_server_class_init(gpointer g_class,
       "SASL mechanisms",
       "The SASL mechanisms offered to the client for authentication",
       NULL,
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
 
@@ -584,7 +617,7 @@ infd_xmpp_server_class_init(gpointer g_class,
       "Security policy",
       "Whether to offer or require TLS",
       INF_TYPE_XMPP_CONNECTION_SECURITY_POLICY,
-      INF_XMPP_CONNECTION_SECURITY_BOTH_PREFER_TLS,
+      INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED,
       G_PARAM_READWRITE | G_PARAM_CONSTRUCT
     )
   );
@@ -661,7 +694,7 @@ infd_xmpp_server_get_type(void)
  * infd_xmpp_server_new:
  * @tcp: A #InfdTcpServer.
  * @policy: The initial security policy.
- * @cred: Certificate credentials used to secure any communication.
+ * @creds: Certificate credentials used to secure any communication.
  * @sasl_context: A SASL context used for authentication.
  * @sasl_mechanisms: A whitespace-sparated list of SASL mechanisms.
  *
@@ -672,9 +705,11 @@ infd_xmpp_server_get_type(void)
  * resulting connection will be in status OPENING until authentication has
  * completed.
  *
- * If @policy is %INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED, then @cred may
- * be %NULL. If @cred is non-%NULL nevertheless, then it is possible to change
+ * If @policy is %INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED, then @creds may
+ * be %NULL. If @creds is non-%NULL nevertheless, then it is possible to change
  * the security policy later using infd_xmpp_server_set_security_policy().
+ * @creds can also be changed later while the server is running. So just set
+ * valid credentials before changing @policy to allow TLS.
  *
  * If @sasl_context is %NULL, the server uses a built-in context that only
  * supports ANONYMOUS authentication. If @sasl_context is not %NULL, then
@@ -687,7 +722,7 @@ infd_xmpp_server_get_type(void)
 InfdXmppServer*
 infd_xmpp_server_new(InfdTcpServer* tcp,
                      InfXmppConnectionSecurityPolicy policy,
-                     gnutls_certificate_credentials_t cred,
+                     InfCertificateCredentials* creds,
                      Gsasl* sasl_context,
                      const gchar* sasl_mechanisms)
 {
@@ -696,15 +731,15 @@ infd_xmpp_server_new(InfdTcpServer* tcp,
   g_return_val_if_fail(INFD_IS_TCP_SERVER(tcp), NULL);
 
   g_return_val_if_fail(
-    policy == INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED || cred != NULL,
+    policy == INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED || creds != NULL,
     NULL
   );
 
   object = g_object_new(
     INFD_TYPE_XMPP_SERVER,
     "tcp-server", tcp,
+    "credentials", creds,
     "security-policy", policy,
-    "credentials", cred,
     "sasl-context", sasl_context,
     "sasl-mechanisms", sasl_mechanisms,
     NULL
