@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinftext/inf-text-chunk.h>
@@ -53,8 +54,11 @@ inf_text_chunk_segment_cmp(gconstpointer first,
                            gconstpointer second,
                            gpointer userdata)
 {
-  const InfTextChunkSegment* first_segment = (InfTextChunkSegment*)first;
-  const InfTextChunkSegment* second_segment = (InfTextChunkSegment*)second;
+  const InfTextChunkSegment* first_segment;
+  const InfTextChunkSegment* second_segment;
+ 
+  first_segment  = (const InfTextChunkSegment*)first;
+  second_segment = (const InfTextChunkSegment*)second;
 
   g_return_val_if_fail(second != NULL && first != NULL, 0);
 
@@ -1128,161 +1132,6 @@ inf_text_chunk_equal(InfTextChunk* self,
   return TRUE;
 }
 
-#if 0
-/**
- * inf_text_chunk_to_xml:
- * @self: A #InfTextChunk.
- * @xml: XML node to write into.
- *
- * Serializes @self into an XML node. This adds &lt;author&gt; child nodes to
- * @xml, with text written by the respective author as content.
- **/
-void
-inf_text_chunk_to_xml(InfTextChunk* self,
-                      xmlNodePtr xml)
-{
-  GSequenceIter* iter;
-  InfTextChunkSegment* segment;
-  gchar* content;
-  gsize utf_bytes;
-  xmlNodePtr child;
-  GIConv cd;
-
-  g_return_if_fail(self != NULL);
-  g_return_if_fail(xml != NULL);
-
-  cd = g_iconv_open("UTF-8", self->encoding);
-  g_assert(cd != (GIConv)-1);
-
-  for(iter = g_sequence_get_begin_iter(self->segments);
-      iter != g_sequence_get_end_iter(self->segments);
-      iter = g_sequence_iter_next(iter))
-  {
-    segment = (InfTextChunkSegment*)g_sequence_get(iter);
-
-    content = g_convert_with_iconv(
-      segment->text,
-      segment->length,
-      cd,
-      NULL,
-      &utf_bytes,
-      NULL
-    );
-
-    /* Conversion into UTF-8 should always succeed */
-    g_assert(content != NULL);
-
-    child = xmlNewTextChild(
-      xml,
-      NULL,
-      (const xmlChar*)"segment",
-      (const xmlChar*)content
-    );
-
-    g_free(content);
-    inf_xml_util_set_attribute_uint(child, "author", segment->author);
-  }
-  
-  g_iconv_close(cd);
-}
-
-/**
- * inf_text_chunk_from_xml:
- * @xml: A XML node.
- * @encoding: Character encoding for the new text chunk.
- * @error: Location to store error information, or %NULL.
- *
- * Converts a XML node created by inf_text_chunk_to_xml() back to a 
- * #InfTextChunk. The created #InfTextChunk will have the given encoding.
- *
- * Return Value: A #InfTextChunk, or %NULL if an error occured.
- **/
-InfTextChunk*
-inf_text_chunk_from_xml(xmlNodePtr xml,
-                        const gchar* encoding,
-                        GError** error)
-{
-  InfTextChunk* chunk;
-  InfTextChunkSegment* segment;
-  guint author;
-  gboolean result;
-
-  xmlNodePtr child;
-  GIConv cd;
-  gsize bytes_written;
-  gsize bytes_read;
-  glong len;
-  guint offset;
-  xmlChar* content;
-  gchar* text;
-
-  g_return_val_if_fail(xml != NULL, NULL);
-  g_return_val_if_fail(encoding != NULL, NULL);
-
-  chunk = inf_text_chunk_new(encoding);
-
-  cd = g_iconv_open(encoding, "UTF-8");
-  g_assert(cd != (GIConv)-1);
-
-  offset = 0;
-
-  for(child = xml->children; child != NULL; child = child->next)
-  {
-    if(strcmp((const char*)child->name, "segment") == 0)
-    {
-      content = xmlNodeGetContent(child);
-      if(content != NULL)
-      {
-        result = inf_xml_util_get_attribute_uint_required(
-          child,
-          "author",
-          &author,
-          error
-        );
-
-        if(result == FALSE)
-        {
-          g_iconv_close(cd);
-          inf_text_chunk_free(chunk);
-          return NULL;
-        }
-
-        text = g_convert_with_iconv(
-          (const gchar*)content,
-          -1,
-          cd,
-          &bytes_read,
-          &bytes_written,
-          error
-        );
-
-        len = g_utf8_strlen(text, bytes_read);
-        xmlFree(content);
-
-        if(text == NULL)
-        {
-          g_iconv_close(cd);
-          inf_text_chunk_free(chunk);
-          return NULL;
-        }
-
-        segment = g_slice_new(InfTextChunkSegment);
-        segment->author = author;
-        segment->text = text;
-        segment->length = bytes_written;
-        segment->offset = offset;
-
-        offset += len;
-        g_sequence_append(chunk->segments, segment);
-      }
-    }
-  }
-  
-  g_iconv_close(cd);
-  return chunk;
-}
-#endif
-
 /**
  * inf_text_chunk_iter_init:
  * @self: A #InfTextChunk.
