@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 #include <libinftext/inf-text-session.h>
@@ -29,6 +30,7 @@
 #include <libinfinity/common/inf-xml-util.h>
 #include <libinfinity/common/inf-error.h>
 #include <libinfinity/inf-i18n.h>
+#include <libinfinity/inf-signals.h>
 
 #include <libxml/tree.h>
 #include <string.h>
@@ -117,7 +119,7 @@ inf_text_session_segment_to_xml(GIConv* cd,
 
   bytes_left = 1024;
 
-  inbuf = (gchar*)text;
+  inbuf = *(gchar**)(gpointer)&text; /* cast const away without warning */
   outbuf = utf8_text;
 
   result = g_iconv(
@@ -131,7 +133,7 @@ inf_text_session_segment_to_xml(GIConv* cd,
   /* Conversion into UTF-8 should always succeed */
   g_assert(result == 0 || errno == E2BIG);
 
-  xmlNodeAddContentLen(xml, (const xmlChar*)utf8_text, 1024 - bytes_left);
+  inf_xml_util_add_child_text(xml, utf8_text, 1024 - bytes_left);
   inf_xml_util_set_attribute_uint(xml, "author", author);
 }
 
@@ -144,27 +146,26 @@ inf_text_session_segment_from_xml(GIConv* cd,
                                   GError** error)
 {
   gsize bytes_read;
-  xmlChar* utf8_text;
+  gchar* utf8_text;
   gpointer text;
 
   if(!inf_xml_util_get_attribute_uint_required(xml, "author", author, error))
     return NULL;
 
-  utf8_text = xmlNodeGetContent(xml);
-  *length = g_utf8_strlen((const gchar*)utf8_text, -1);
+  utf8_text = inf_xml_util_get_child_text(xml, &bytes_read, length, error);
+  if(!utf8_text)
+    return NULL;
 
   text = g_convert_with_iconv(
-    (const gchar*)utf8_text,
-    -1,
+    utf8_text,
+    bytes_read,
     *cd,
-    &bytes_read,
+    NULL,
     bytes,
     error
   );
 
-  xmlFree(utf8_text);
-  if(text == NULL) return FALSE;
-
+  g_free(utf8_text);
   return text;
 }
 
@@ -263,6 +264,7 @@ inf_text_session_selection_changed_cb(InfTextUser* user,
   session = INF_TEXT_SESSION(user_data);
   priv = INF_TEXT_SESSION_PRIVATE(session);
   local = inf_text_session_find_local_user(session, user);
+  g_assert(local != NULL);
 
   g_get_current_time(&current);
   diff = inf_text_session_timeval_diff(&current, &local->last_caret_update);
@@ -329,6 +331,12 @@ inf_text_session_remove_local_user(InfTextSession* session,
     );
   }
 
+  inf_signal_handlers_disconnect_by_func(
+    G_OBJECT(local->user),
+    G_CALLBACK(inf_text_session_selection_changed_cb),
+    session
+  );
+
   g_slice_free(InfTextSessionLocalUser, local);
   priv->local_users = g_slist_remove(priv->local_users, local);
 }
@@ -376,7 +384,7 @@ inf_text_session_block_local_users_selection_changed(InfTextSession* session)
   {
     local = (InfTextSessionLocalUser*)item->data;
 
-    g_signal_handlers_block_by_func(
+    inf_signal_handlers_block_by_func(
       G_OBJECT(local->user),
       G_CALLBACK(inf_text_session_selection_changed_cb),
       session
@@ -397,7 +405,7 @@ inf_text_session_unblock_local_users_selection_changed(InfTextSession* sess)
   {
     local = (InfTextSessionLocalUser*)item->data;
 
-    g_signal_handlers_unblock_by_func(
+    inf_signal_handlers_unblock_by_func(
       G_OBJECT(local->user),
       G_CALLBACK(inf_text_session_selection_changed_cb),
       sess
@@ -633,13 +641,13 @@ inf_text_session_apply_request_cb_before(InfAdoptedAlgorithm* algorithm,
   session = INF_TEXT_SESSION(user_data);
   buffer = INF_TEXT_BUFFER(inf_session_get_buffer(INF_SESSION(session)));
 
-  g_signal_handlers_block_by_func(
+  inf_signal_handlers_block_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_insert_text_cb_before),
     session
   );
 
-  g_signal_handlers_block_by_func(
+  inf_signal_handlers_block_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_erase_text_cb_before),
     session
@@ -660,13 +668,13 @@ inf_text_session_apply_request_cb_after(InfAdoptedAlgorithm* algorithm,
   session = INF_TEXT_SESSION(user_data);
   buffer = INF_TEXT_BUFFER(inf_session_get_buffer(INF_SESSION(session)));
 
-  g_signal_handlers_unblock_by_func(
+  inf_signal_handlers_unblock_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_insert_text_cb_before),
     session
   );
   
-  g_signal_handlers_unblock_by_func(
+  inf_signal_handlers_unblock_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_erase_text_cb_before),
     session
@@ -808,11 +816,11 @@ inf_text_session_constructor(GType type,
    * empty (we will fill it during synchronization). Text handlers are
    * connected when synchronization is complete. */
   g_assert(
-    status != INF_SESSION_SYNCHRONIZING ||
+    status == INF_SESSION_RUNNING ||
     inf_text_buffer_get_length(buffer) == 0
   );
 
-  if(status != INF_SESSION_SYNCHRONIZING)
+  if(status == INF_SESSION_RUNNING)
     inf_text_session_init_text_handlers(session);
 
   return object;
@@ -859,37 +867,37 @@ inf_text_session_dispose(GObject* object)
     );
   }
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_insert_text_cb_before),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_erase_text_cb_before),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_insert_text_cb_after),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(buffer),
     G_CALLBACK(inf_text_session_buffer_erase_text_cb_after),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(user_table),
     G_CALLBACK(inf_text_session_local_user_added_cb),
     session
   );
 
-  g_signal_handlers_disconnect_by_func(
+  inf_signal_handlers_disconnect_by_func(
     G_OBJECT(user_table),
     G_CALLBACK(inf_text_session_local_user_removed_cb),
     session
@@ -898,13 +906,13 @@ inf_text_session_dispose(GObject* object)
   /* Algorithm may be NULL if we were still synchronizing */
   if(algorithm != NULL)
   {
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(algorithm),
       G_CALLBACK(inf_text_session_apply_request_cb_before),
       session
     );
 
-    g_signal_handlers_disconnect_by_func(
+    inf_signal_handlers_disconnect_by_func(
       G_OBJECT(algorithm),
       G_CALLBACK(inf_text_session_apply_request_cb_after),
       session
@@ -944,7 +952,7 @@ inf_text_session_set_property(GObject* object,
     priv->caret_update_interval = g_value_get_uint(value);
     break;
   default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID(value, prop_id, pspec);
+    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
     break;
   }
 }
@@ -1018,6 +1026,7 @@ inf_text_session_handle_user_color_change(InfTextSession* session,
       error,
       inf_user_error_quark(),
       INF_USER_ERROR_NOT_JOINED,
+      "%s",
       _("User did not join from this connection")
     );
 
@@ -1350,8 +1359,9 @@ inf_text_session_validate_user_props(InfSession* session,
   {
     g_set_error(
       error,
-      inf_text_session_error_quark,
+      inf_request_error_quark(),
       INF_REQUEST_ERROR_NO_SUCH_ATTRIBUTE,
+      "%s",
       _("'caret' attribute in user message is missing")
     );
 
@@ -1462,7 +1472,7 @@ inf_text_session_request_to_xml(InfAdoptedSession* session,
       g_assert(utf8_text != NULL);
       g_assert(bytes_read == inf_text_chunk_iter_get_bytes(&iter));
 
-      xmlNodeAddContentLen(op_xml, (const xmlChar*)utf8_text, bytes_written);
+      inf_xml_util_add_child_text(op_xml, utf8_text, bytes_written);
       g_free(utf8_text);
 
       /* We only allow a single segment because the whole inserted text must
@@ -1494,8 +1504,6 @@ inf_text_session_request_to_xml(InfAdoptedSession* session,
         cd = g_iconv_open("UTF-8", inf_text_chunk_get_encoding(chunk));
         result = inf_text_chunk_iter_init(chunk, &iter);
 
-        /* Should delete something, otherwise we could also use noop */
-        g_assert(result == TRUE);
         while(result == TRUE)
         {
           text = inf_text_chunk_iter_get_text(&iter);
@@ -1602,10 +1610,11 @@ inf_text_session_xml_to_request(InfAdoptedSession* session,
   guint pos;
   gchar* text;
   gsize bytes;
-  guint length;
   InfTextChunk* chunk;
 
-  xmlChar* utf8_text;
+  gchar* utf8_text;
+  gsize in_bytes;
+  guint length;
 
   xmlNodePtr child;
   GIConv cd;
@@ -1637,15 +1646,13 @@ inf_text_session_xml_to_request(InfAdoptedSession* session,
     if(!inf_xml_util_get_attribute_uint_required(op_xml, "pos", &pos, error))
       goto fail;
 
-    /* TODO: Use XML_GET_CONTENT to avoid copy? */
-    /* TODO: Can we find out character and byte count in one pass?
-     * g_convert() calls strlen. */
-    utf8_text = xmlNodeGetContent(op_xml);
-    length = g_utf8_strlen((const gchar*)utf8_text, -1);
+    utf8_text = inf_xml_util_get_child_text(op_xml, &in_bytes, &length, error);
+    if(!utf8_text)
+      goto fail;
 
     text = g_convert(
-      (const gchar*)utf8_text,
-      -1,
+      utf8_text,
+      in_bytes,
       inf_text_buffer_get_encoding(buffer),
       "UTF-8",
       NULL,
@@ -1653,7 +1660,7 @@ inf_text_session_xml_to_request(InfAdoptedSession* session,
       error
     );
 
-    xmlFree(utf8_text);
+    g_free(utf8_text);
     if(text == NULL) goto fail;
 
     chunk = inf_text_chunk_new(inf_text_buffer_get_encoding(buffer));
@@ -1913,19 +1920,21 @@ inf_text_session_get_type(void)
  * @manager: A #InfCommunicationManager.
  * @buffer: An initial #InfTextBuffer.
  * @io: A #InfIo object.
- * @sync_group: A group in which the session is synchronized, or %NULL.
+ * @status: The initial status of the session.
+ * @sync_group: A group in which the session is synchronized. Ignored if
+ * @status is %INF_SESSION_RUNNING.
  * @sync_connection: A connection to synchronize the session from. Ignored if
- * @sync_group is %NULL.
+ * @status is %INF_SESSION_RUNNING.
  *
  * Creates a new #InfTextSession. The communication manager is used to send
  * and receive requests from subscription and synchronization. @buffer will be
  * set to be initially empty if the session is initially synchronized
  * (see below). @io is required to trigger timeouts.
  *
- * If @sync_group is not %NULL, the session will initially be sychronized,
- * meaning the initial content is retrieved from @sync_connection. If you are
- * subscribed to the session, set the subscription group via
- * inf_session_set_subscription_group().
+ * If @status is %INF_SESSION_PRESYNC or %INF_SESSION_SYNCHRONIZING, then the
+ * session will initially be sychronized, meaning the initial content is
+ * retrieved from @sync_connection. If you are subscribed to the session, set
+ * the subscription group via inf_session_set_subscription_group().
  *
  * Return Value: A new #InfTextSession.
  **/
@@ -1933,6 +1942,7 @@ InfTextSession*
 inf_text_session_new(InfCommunicationManager* manager,
                      InfTextBuffer* buffer,
                      InfIo* io,
+                     InfSessionStatus status,
                      InfCommunicationGroup* sync_group,
                      InfXmlConnection* sync_connection)
 {
@@ -1943,8 +1953,10 @@ inf_text_session_new(InfCommunicationManager* manager,
   g_return_val_if_fail(INF_IS_IO(io), NULL);
 
   g_return_val_if_fail(
-    sync_group == NULL || 
-    (INF_COMMUNICATION_IS_GROUP(sync_group) &&
+    (status == INF_SESSION_RUNNING &&
+     sync_group == NULL && sync_connection == NULL) ||
+    (status != INF_SESSION_RUNNING &&
+     INF_COMMUNICATION_IS_GROUP(sync_group) &&
      INF_IS_XML_CONNECTION(sync_connection)),
     NULL
   );
@@ -1953,6 +1965,7 @@ inf_text_session_new(InfCommunicationManager* manager,
     INF_TEXT_TYPE_SESSION,
     "communication-manager", manager,
     "buffer", buffer,
+    "status", status,
     "sync-group", sync_group,
     "sync-connection", sync_connection,
     "io", io,
@@ -1968,19 +1981,21 @@ inf_text_session_new(InfCommunicationManager* manager,
  * @buffer: An initial #InfTextBuffer.
  * @io: A #InfIo object.
  * @user_table: A #InfUserTable.
- * @sync_group: A group in which the session is synchronized, or %NULL.
+ * @status: The initial status for the session.
+ * @sync_group: A group in which the session is synchronized. Ignored if
+ * @status is %INF_SESSION_RUNNING.
  * @sync_connection: A connection to synchronize the session from. Ignored if
- * @sync_group is %NULL.
+ * @status is %INF_SESSION_RUNNING.
  *
  * Creates a new #InfTextSession. The connection manager is used to send and
  * receive requests from subscription and synchronization. @buffer will be
  * set to be initially empty if the session is initially synchronized
  * (see below). @io is required to trigger timeouts.
  *
- * If @sync_group is not %NULL, the session will initially be sychronized,
- * meaning the initial content is retrieved from @sync_connection. If you are
- * subscribed to the session, set the subscription group via
- * inf_session_set_subscription_group().
+ * If @status is %INF_SESSION_PRESYNC or %INF_SESSION_SYNCHRONIZING, then the
+ * session will initially be sychronized, meaning the initial content is
+ * retrieved from @sync_connection. If you are subscribed to the session, set
+ * the subscription group via inf_session_set_subscription_group().
  *
  * @user_table is used as an initial user table. The user table should only
  * contain unavailable users, if any, that may rejoin during the session. If
@@ -1995,9 +2010,13 @@ inf_text_session_new_with_user_table(InfCommunicationManager* manager,
                                      InfTextBuffer* buffer,
                                      InfIo* io,
                                      InfUserTable* user_table,
+                                     InfSessionStatus status,
                                      InfCommunicationGroup* sync_group,
                                      InfXmlConnection* sync_connection)
 {
+  /* TODO: Can it happen that the user_table is set explicitely PLUS the
+   * session is synchronized? If not then this function can be simplified */
+
   GObject* object;
 
   g_return_val_if_fail(INF_COMMUNICATION_IS_MANAGER(manager), NULL);
@@ -2006,8 +2025,10 @@ inf_text_session_new_with_user_table(InfCommunicationManager* manager,
   g_return_val_if_fail(INF_IS_USER_TABLE(user_table), NULL);
 
   g_return_val_if_fail(
-    sync_group == NULL ||
-    (INF_COMMUNICATION_IS_GROUP(sync_group) &&
+    (status == INF_SESSION_RUNNING &&
+     sync_group == NULL && sync_connection == NULL) ||
+    (status != INF_SESSION_RUNNING &&
+     INF_COMMUNICATION_IS_GROUP(sync_group) &&
      INF_IS_XML_CONNECTION(sync_connection)),
     NULL
   );
@@ -2017,6 +2038,7 @@ inf_text_session_new_with_user_table(InfCommunicationManager* manager,
     "communication-manager", manager,
     "buffer", buffer,
     "user-table", user_table,
+    "status", status,
     "sync-group", sync_group,
     "sync-connection", sync_connection,
     "io", io,
@@ -2057,7 +2079,7 @@ inf_text_session_set_user_color(InfTextSession* session,
   inf_xml_util_set_attribute_uint(xml, "id", inf_user_get_id(INF_USER(user)));
   inf_xml_util_set_attribute_double(xml, "hue", hue);
 
-  inf_session_send_to_subscriptions(INF_SESSION(session), NULL, xml);
+  inf_session_send_to_subscriptions(INF_SESSION(session), xml);
   g_object_set(G_OBJECT(user), "hue", hue, NULL);
 }
 
@@ -2075,7 +2097,8 @@ inf_text_session_set_user_color(InfTextSession* session,
  * are automatically activated as soon as they issue a request.
  *
  * TODO: We should probably detect this automatically, without requiring
- * people to call this function.
+ * people to call this function, i.e. flush requests for local users just
+ * before they become inactive.
  *
  * @user must have the %INF_USER_LOCAL flag set.
  */
