@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
  */
 
 /* TODO: Add some of the replay logic as public libinfinity API */
@@ -21,6 +22,7 @@
 #include "util/inf-test-util.h"
 
 #include <libinftext/inf-text-session.h>
+#include <libinftext/inf-text-undo-grouping.h>
 #include <libinftext/inf-text-default-buffer.h>
 #include <libinftext/inf-text-default-insert-operation.h>
 #include <libinftext/inf-text-insert-operation.h>
@@ -30,6 +32,7 @@
 #include <libinfinity/communication/inf-communication-manager.h>
 #include <libinfinity/common/inf-simulated-connection.h>
 #include <libinfinity/common/inf-standalone-io.h>
+#include <libinfinity/common/inf-xml-util.h>
 #include <libinfinity/common/inf-io.h>
 
 #include <libxml/xmlreader.h>
@@ -50,6 +53,13 @@ typedef enum _InfTestTextReplayError {
   INF_TEST_TEXT_REPLAY_UNEXPECTED_NODE
 } InfTestTextReplayError;
 
+typedef struct _InfTestTextReplayPlayUserTableForeachFuncData
+  InfTestTextReplayPlayUserTableForeachFuncData;
+struct _InfTestTextReplayPlayUserTableForeachFuncData {
+  InfAdoptedAlgorithm* algorithm;
+  GSList* undo_groupings;
+};
+
 typedef struct _InfTestTextReplayObject InfTestTextReplayObject;
 struct _InfTestTextReplayObject {
   GObject parent_instance;
@@ -94,8 +104,7 @@ G_DEFINE_TYPE_WITH_CODE(
   )
 )
 
-/* TODO: These functions assume that the buffer contains ASCII-only text.
- * It will fail with UTF-8! */
+/* The next three functions assume that buffer and chunks contain UTF-8 */
 static GString*
 inf_test_text_replay_load_buffer(InfTextBuffer* buffer)
 {
@@ -130,6 +139,7 @@ inf_test_text_replay_apply_operation_to_string(GString* string,
   InfTextChunk* chunk;
   InfTextChunkIter iter;
   guint position;
+  guint length;
 
   if(INF_TEXT_IS_INSERT_OPERATION(operation))
   {
@@ -145,6 +155,9 @@ inf_test_text_replay_apply_operation_to_string(GString* string,
 
     if(inf_text_chunk_iter_init(chunk, &iter))
     {
+      /* Convert from pos to byte */
+      position = g_utf8_offset_to_pointer(string->str, position) - string->str;
+
       do
       {
         g_string_insert_len(
@@ -160,15 +173,19 @@ inf_test_text_replay_apply_operation_to_string(GString* string,
   }
   else if(INF_TEXT_IS_DELETE_OPERATION(operation))
   {
-    g_string_erase(
-      string,
-      inf_text_delete_operation_get_position(
-        INF_TEXT_DELETE_OPERATION(operation)
-      ),
-      inf_text_delete_operation_get_length(
-        INF_TEXT_DELETE_OPERATION(operation)
-      )
+    position = inf_text_delete_operation_get_position(
+      INF_TEXT_DELETE_OPERATION(operation)
     );
+    length = inf_text_delete_operation_get_length(
+      INF_TEXT_DELETE_OPERATION(operation)
+    );
+
+    length = g_utf8_offset_to_pointer(string->str, position+length) -
+      string->str;
+    position = g_utf8_offset_to_pointer(string->str, position) - string->str;
+    length -= position;
+
+    g_string_erase(string, position, length);
   }
 }
 
@@ -380,8 +397,10 @@ inf_test_text_replay_play_initial(xmlTextReaderPtr reader,
 static gboolean
 inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
                                    InfSession* session,
+                                   GSList** undo_groupings,
                                    InfCommunicationGroup* publisher_group,
                                    InfXmlConnection* publisher,
+                                   InfXmlConnection* client,
                                    GError** error)
 {
   InfSessionClass* session_class;
@@ -389,6 +408,9 @@ inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
   xmlNodePtr cur;
   InfUser* user;
   guint i;
+  InfTextUndoGrouping* grouping;
+  GParameter* param;
+  guint user_id;
 
   session_class = INF_SESSION_GET_CLASS(session);
 
@@ -399,9 +421,29 @@ inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
 
     if(strcmp((const char*)cur->name, "request") == 0)
     {
+      /* TODO: Add user join/leaves to record. */
+      /* Until that is done, make users available when they issue a request */
+      if(!inf_xml_util_get_attribute_uint_required(cur, "user", &user_id, error))
+        return FALSE;
+
+      user = inf_user_table_lookup_user_by_id(
+        inf_session_get_user_table(session),
+        user_id
+      );
+      g_assert(user);
+
+      if(inf_user_get_status(user) == INF_USER_UNAVAILABLE)
+      {
+        g_object_set(
+          G_OBJECT(user),
+          "status", INF_USER_ACTIVE,
+          "connection", client,
+          NULL
+        );
+      }
+
       inf_communication_group_send_group_message(
         publisher_group,
-        NULL,
         xmlCopyNode(cur, 1)
       );
 
@@ -412,6 +454,12 @@ inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
     {
       /* User join */
       user_props = session_class->get_xml_user_props(session, publisher, cur);
+      param = inf_session_get_user_property(user_props, "connection");
+      if(!G_IS_VALUE(&param->value))
+      {
+        g_value_init(&param->value, INF_TYPE_XML_CONNECTION);
+        g_value_set_object(&param->value, G_OBJECT(client));
+      }
 
       user = inf_session_add_user(
         session,
@@ -423,6 +471,15 @@ inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
       for(i = 0; i < user_props->len; ++ i)
         g_value_unset(&g_array_index(user_props, GParameter, i).value);
 
+      grouping = inf_text_undo_grouping_new();
+      inf_adopted_undo_grouping_set_algorithm(
+        INF_ADOPTED_UNDO_GROUPING(grouping),
+        inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session)),
+        INF_ADOPTED_USER(user)
+      );
+
+      *undo_groupings = g_slist_prepend(*undo_groupings, grouping);
+
       g_array_free(user_props, TRUE);
       if(user == NULL) return FALSE;
     }
@@ -449,11 +506,28 @@ inf_test_text_replay_play_requests(xmlTextReaderPtr reader,
   return TRUE;
 }
 
+static void
+inf_test_text_replay_play_user_table_foreach_func(InfUser* user,
+                                                  gpointer user_data)
+{
+  InfTextUndoGrouping* grouping;
+  InfTestTextReplayPlayUserTableForeachFuncData* data;
+
+  grouping = inf_text_undo_grouping_new();
+  data = (InfTestTextReplayPlayUserTableForeachFuncData*)user_data;
+
+  inf_adopted_undo_grouping_set_algorithm(INF_ADOPTED_UNDO_GROUPING(grouping),
+                                          data->algorithm,
+                                          INF_ADOPTED_USER(user));
+  data->undo_groupings = g_slist_prepend(data->undo_groupings, grouping);
+}
+
 static gboolean
 inf_test_text_replay_play(xmlTextReaderPtr reader,
                           InfSession* session,
                           InfCommunicationGroup* publisher_group,
                           InfXmlConnection* publisher,
+                          InfXmlConnection* client,
                           GError** error)
 {
   gboolean result;
@@ -461,6 +535,9 @@ inf_test_text_replay_play(xmlTextReaderPtr reader,
   /* Used to find InfTextChunk errors */
   GString* content;
 
+  InfUserTable* user_table;
+  InfTestTextReplayPlayUserTableForeachFuncData data;
+
   /* Advance to root node */
   if(xmlTextReaderNodeType(reader) != XML_READER_TYPE_ELEMENT)
     if(inf_test_text_replay_advance_required(reader, error) == FALSE)
@@ -539,14 +616,35 @@ inf_test_text_replay_play(xmlTextReaderPtr reader,
     content
   );
 
+  /* Let an undo grouper group stuff, just as a consistency check
+   * that it does not crash or behave badly. */
+  user_table = inf_session_get_user_table(session);
+  data.algorithm =
+    inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session));
+  data.undo_groupings = NULL;
+  inf_user_table_foreach_user(
+    user_table,
+    inf_test_text_replay_play_user_table_foreach_func,
+    &data
+  );
+
   result = inf_test_text_replay_play_requests(
     reader,
     session,
+    &data.undo_groupings,
     publisher_group,
     publisher,
+    client,
     error
   );
 
+  while(data.undo_groupings != NULL)
+  {
+    g_object_unref(data.undo_groupings->data);
+    data.undo_groupings =
+      g_slist_remove(data.undo_groupings, data.undo_groupings->data);
+  }
+
   g_string_free(content, TRUE);
 
   if(!result) return FALSE;
@@ -633,6 +731,7 @@ inf_test_text_replay_process(xmlTextReaderPtr reader)
     client_manager,
     buffer,
     io,
+    INF_SESSION_SYNCHRONIZING,
     INF_COMMUNICATION_GROUP(client_group),
     INF_XML_CONNECTION(client)
   );
@@ -659,6 +758,7 @@ inf_test_text_replay_process(xmlTextReaderPtr reader)
     INF_SESSION(session),
     INF_COMMUNICATION_GROUP(publisher_group),
     INF_XML_CONNECTION(publisher),
+    INF_XML_CONNECTION(client),
     &error
   );
 
